from numbers import Number

from _common_decl cimport *
from cython.operator cimport dereference as deref

from _cy_affine cimport cy_Translate, cy_Rotate, cy_Scale
from _cy_affine cimport cy_VShear, cy_HShear, cy_Zoom
from _cy_affine cimport cy_Affine, get_Affine, is_transform

from _cy_curves cimport cy_Curve, wrap_Curve_p
from _cy_curves cimport cy_LineSegment, wrap_LineSegment

cdef class cy_Angle:

    """Class representig angle.

    Angles can be tested for equality, but they are not ordered.

    Corresponds to Angle class in 2geom. Most members are direct
    calls to Angle methods, otherwise C++ call is specified.
    """

    def __cinit__(self, double x):
        """Create new angle from value in radians."""
        self.thisptr = new Angle(x)

    def __repr__(self):
        """repr(self)"""
        return "Angle({0:2.f})".format(self.radians())

    def __str__(self):
        """str(self)"""
        return "{0:.2f} radians".format(self.radians())

    def __dealloc__(self):
        del self.thisptr

    @classmethod
    def from_radians(cls, rad):
        """Construnct angle from radians."""
        return wrap_Angle(from_radians(rad))

    @classmethod
    def from_degrees(cls, d):
        """Construnct angle from degrees."""
        return wrap_Angle(from_degrees(d))

    @classmethod
    def from_degrees_clock(cls, d):
        """Construnct angle from degrees in clock convention."""
        return wrap_Angle(from_degrees_clock(d))

    @classmethod
    def from_Point(cls, cy_Point p):
        """Construct angle from Point. Calls Angle(Point) in 2geom."""
        cdef Point * pp = p.thisptr
        return wrap_Angle( Angle( deref(p.thisptr) ))

    def __float__(self):
        """float(self)"""
        return <Coord> deref(self.thisptr)

    def __add__(cy_Angle self, cy_Angle other):
        """alpha + beta"""
        return wrap_Angle(deref(other.thisptr) + deref(self.thisptr))

    def __sub__(cy_Angle self, cy_Angle other):
        """alpha - beta"""
        return wrap_Angle(deref(other.thisptr) - deref(self.thisptr))

    def __richcmp__(cy_Angle self, cy_Angle other, int op):
        """Test equality of angles. Note: angles are not ordered."""
        if op == 2:
            return deref(other.thisptr) == deref(self.thisptr)
        elif op == 3:
            return deref(other.thisptr) != deref(self.thisptr)
        return NotImplemented

    def radians(self):
        """Return the angle in radians."""
        return self.thisptr.radians()

    def radians0(self):
        """Return the angle in positive radians."""
        return self.thisptr.radians0()

    def degrees(self):
        """Return the angle in degrees."""
        return self.thisptr.degrees()

    def degrees_clock(self):
        """Return the angle in clock convention. Calls degreesClock() in 2geom."""
        return self.thisptr.degreesClock()

    @classmethod
    def rad_from_deg(cls, deg):
        """Convert degrees to radians."""
        return rad_from_deg(deg)

    @classmethod
    def deg_from_rad(cls, rad):
        """Convert radians to degrees."""
        return deg_from_rad(rad)

cdef cy_Angle wrap_Angle(Angle p):
    cdef Angle * retp = new Angle()
    retp[0] = p
    cdef cy_Angle r = cy_Angle.__new__(cy_Angle, 0)
    r.thisptr = retp
    return r


cdef class cy_AngleInterval:

    """ Class representing interval of angles.

    Corresponds to AngleInterval class in 2geom. Most members are direct
    calls to AngleInterval methods, otherwise C++ call is specified.
    """

    cdef AngleInterval* thisptr

    def __cinit__(self, start, end, bint cw=False):
        """Create AngleInterval from starting and ending value.

        Optional argument cw specifies direction - counter-clockwise
        is default.
        """
        self.thisptr = new AngleInterval(float(start), float(end), cw)

    def __call__(self, Coord t):
        """A(t), maps unit interval to Angle."""
        return wrap_Angle(deref( self.thisptr ) (t))

    def initial_angle(self):
        """Return initial angle as Angle instance."""
        return wrap_Angle(self.thisptr.initialAngle())

    def final_angle(self):
        """Return final angle as Angle instance."""
        return wrap_Angle(self.thisptr.finalAngle())

    def is_degenerate(self):
        """Test for empty interval."""
        return self.thisptr.isDegenerate()

    def angle_at(self, Coord t):
        """A.angle_at(t) <==> A(t)"""
        return wrap_Angle(self.thisptr.angleAt(t))

    def contains(self, cy_Angle a):
        """Test whether interval contains Angle."""
        return self.thisptr.contains(deref( a.thisptr ))

    def extent(self):
        """Calculate interval's extent."""
        return self.thisptr.extent()

cdef cy_AngleInterval wrap_AngleInterval(AngleInterval p):
    cdef AngleInterval * retp = new AngleInterval(0, 0, 0)
    retp[0] = p
    cdef cy_AngleInterval r = cy_AngleInterval.__new__(cy_AngleInterval)
    r.thisptr = retp
    return r


cdef class cy_Point:

    """Represents point or vector in 2D plane.

    Points are ordered lexicographically, with y coordinate being
    more significant.

    Corresponds to Point class in 2geom. Most members are direct
    calls to Point methods, otherwise C++ call is specified.
    """

    def __cinit__(self, double x=0.0, double y=0.0):
        """Create Point from it's cartesian coordinates."""
        self.thisptr = new Point(x, y)

    def __repr__(self):
        """repr(self)"""
        return "Point ({0:.3f}, {1:.3f})".format(self[0], self[1])

    def __str__(self):
        """str(self)"""
        return "[{0:.3f}, {1:.3f}]".format(self[0], self[1])

    def __dealloc__(self):
        del self.thisptr

    @classmethod
    def polar(cls, Coord angle, Coord radius = 1.0):
        """Create Point from polar coordinates."""
        return wrap_Point(polar(angle, radius))

    def length(self):
        """Return distance from origin or length of vector."""
        return self.thisptr.length()

    def ccw(self):
        """Return point rotated counter-clockwise."""
        return wrap_Point(self.thisptr.ccw())

    def cw(self):
        """Return point rotated clockwise."""
        return wrap_Point(self.thisptr.cw())

    def __getitem__(self, key):
        """Access coordinates of point."""
        return deref(self.thisptr)[key]

    @property
    def x(self):
        """First coordinate of point."""
        return self.thisptr.x()

    @property
    def y(self):
        """Second coordinate of point."""
        return self.thisptr.y()

    def round(self):
        """Create IntPoint rounding coordinates."""
        return wrap_IntPoint(self.thisptr.round())

    def floor(self):
        """Create IntPoint flooring coordinates."""
        return wrap_IntPoint(self.thisptr.floor())

    def ceil(self):
        """Create IntPoint ceiling coordinates."""
        return wrap_IntPoint(self.thisptr.ceil())

    def __neg__(self):
        """-P"""
        return(wrap_Point(-deref(self.thisptr)))

    def __abs__(self):
        """abs(P)"""
        return self.length()

    def __add__(cy_Point self, cy_Point other):
        """P + Q"""
        return wrap_Point(deref(self.thisptr) + deref(other.thisptr))

    def __sub__(cy_Point self, cy_Point other):
        """P - Q"""
        return wrap_Point(deref(self.thisptr) - deref(other.thisptr))

    #TODO exceptions
    def __mul__(cy_Point self, s):
        """Multiply point by number or Affine transform."""
        if isinstance(s, Number):
            return wrap_Point(deref(self.thisptr)* (<Coord> float(s)))
        elif isinstance(s, cy_Affine):
            return wrap_Point( deref(self.thisptr) * <Affine &> deref( (<cy_Affine> s).thisptr ) )
        elif isinstance(s, cy_Translate):
            return wrap_Point( deref(self.thisptr) * <Translate &> deref( (<cy_Translate> s).thisptr ) )
        elif isinstance(s, cy_Scale):
            return wrap_Point( deref(self.thisptr) * <Scale &> deref( (<cy_Scale> s).thisptr ) )
        elif isinstance(s, cy_Rotate):
            return wrap_Point( deref(self.thisptr) * <Rotate &> deref( (<cy_Rotate> s).thisptr ) )
        elif isinstance(s, cy_HShear):
            return wrap_Point( deref(self.thisptr) * <HShear &> deref( (<cy_HShear> s).thisptr ) )
        elif isinstance(s, cy_VShear):
            return wrap_Point( deref(self.thisptr) * <VShear &> deref( (<cy_VShear> s).thisptr ) )
        elif isinstance(s, cy_Zoom):
            return wrap_Point( deref(self.thisptr) * <Zoom &> deref( (<cy_Zoom> s).thisptr ) )
        return NotImplemented

    def __div__(cy_Point self, Coord s):
        """P / s"""
        return wrap_Point(deref(self.thisptr)/s)

    def __richcmp__(cy_Point self, cy_Point other, int op):
        if op == 0:
            return deref(self.thisptr) < deref(other.thisptr)
        if op == 1:
            return deref(self.thisptr) <= deref(other.thisptr)
        if op == 2:
            return deref(self.thisptr) == deref(other.thisptr)
        if op == 3:
            return deref(self.thisptr) != deref(other.thisptr)
        if op == 4:
            return deref(self.thisptr) > deref(other.thisptr)
        if op == 5:
            return deref(self.thisptr) >= deref(other.thisptr)

    def isFinite(self):
        """Test whether point is finite."""
        return self.thisptr.isFinite()

    def isZero(self):
        """Test whether point is origin"""
        return self.thisptr.isZero()

    def isNormalized(self, eps=EPSILON):
        """Test whether point's norm is close to 1."""
        return self.thisptr.isNormalized(eps)

    @classmethod
    def L2(cls, cy_Point p):
        """Compute L2 (Euclidean) norm of point.

        L2(P) = sqrt( P.x**2 + P.y**2 )
        """
        return L2(deref(p.thisptr))

    @classmethod
    def L2sq(cls, cy_Point p):
        """Compute square of L2 (Euclidean) norm."""
        return L2sq(deref(p.thisptr))

    @classmethod
    def are_near(cls, cy_Point a, cy_Point b, double eps=EPSILON):
        """Test if two points are close."""
        return are_near(deref(a.thisptr), deref(b.thisptr), eps)

    @classmethod
    def middle_point(cls, cy_Point a, cy_Point b):
        """Return point between two points."""
        return wrap_Point(middle_point(deref(a.thisptr), deref(b.thisptr)))

    @classmethod
    def rot90(cls, cy_Point a):
        """Rotate point by 90 degrees."""
        return wrap_Point(rot90(deref(a.thisptr)))

    @classmethod
    def lerp(cls, double t, cy_Point a, cy_Point b):
        """Linearly interpolate between too points."""
        return wrap_Point(lerp(t, deref(a.thisptr), deref(b.thisptr)))

    @classmethod
    def dot(cls, cy_Point a, cy_Point b):
        """Calculate dot product of two points."""
        return dot(deref(a.thisptr), deref(b.thisptr))

    @classmethod
    def cross(cls, cy_Point a, cy_Point b):
        """Calculate (z-coordinate of) cross product of two points."""
        return cross(deref(a.thisptr), deref(b.thisptr))

    @classmethod
    def distance(cls, cy_Point a, cy_Point b):
        """Compute distance between two points."""
        return distance(deref(a.thisptr), deref(b.thisptr))

    @classmethod
    def distanceSq(cls, cy_Point a, cy_Point b):
        """Compute square of distance between two points."""
        return distanceSq(deref(a.thisptr), deref(b.thisptr))

    @classmethod
    def unit_vector(cls, cy_Point p):
        """Normalise point."""
        return wrap_Point(unit_vector(deref(p.thisptr)))

    @classmethod
    def L1(cls, cy_Point p):
        """Compute L1 (Manhattan) norm of a point.

        L1(P) = |P.x| + |P.y|
        """
        return L1(deref(p.thisptr))

    @classmethod
    def LInfty(cls, cy_Point p):
        """Compute Infinity norm of a point.

        LInfty(P) = max(|P.x|, |P.y|)
        """
        return LInfty(deref(p.thisptr))

    @classmethod
    def is_zero(cls, cy_Point p):
        """Test whether point is origin."""
        return is_zero(deref(p.thisptr))

    @classmethod
    def is_unit_vector(cls, cy_Point p):
        """Test whether point's length equal 1."""
        return is_unit_vector(deref(p.thisptr))

    @classmethod
    def atan2(cls, cy_Point p):
        """Return angle between point and x-axis."""
        return atan2(deref(p.thisptr))

    @classmethod
    def angle_between(cls, cy_Point a, cy_Point b):
        """Return angle between two point."""
        return angle_between(deref(a.thisptr), deref(b.thisptr))

    @classmethod
    def abs(cls, cy_Point p):
        """Return length of a point."""
        return wrap_Point(abs(deref(p.thisptr)))

    @classmethod
    def constrain_angle(cls, cy_Point a, cy_Point b, unsigned int n, cy_Point direction):
        """Rotate B around A to have specified angle wrt. direction."""
        return wrap_Point(constrain_angle(deref(a.thisptr), deref(b.thisptr), n, deref(direction.thisptr)))

cdef cy_Point wrap_Point(Point p):
    cdef Point * retp = new Point()
    retp[0] = p
    cdef cy_Point r = cy_Point.__new__(cy_Point)
    r.thisptr = retp
    return r

cdef object wrap_vector_point(vector[Point] v):
    r = []
    cdef unsigned int i
    for i in range(v.size()):
        r.append( wrap_Point(v[i]) )
    return r

cdef vector[Point] make_vector_point(object l):
    cdef vector[Point] ret
    for i in l:
        ret.push_back( deref( (<cy_Point> i).thisptr ) )
    return ret


cdef class cy_IntPoint:

    """Represents point with integer coordinates

    IntPoints are ordered lexicographically, with y coordinate being
    more significant.

    Corresponds to IntPoint class in 2geom. Most members are direct
    calls to IntPoint methods, otherwise C++ call is specified.
    """

    def __init__(self, IntCoord x = 0, IntCoord y = 0):
        """Create new IntPoint from it's cartesian coordinates."""
        self.thisptr =  new IntPoint(x ,y)

    def __getitem__(self, key):
        """Get coordinates of IntPoint."""
        return deref(self.thisptr)[key]

    def __repr__(self):
        """repr(self)"""
        return "IntPoint ({0}, {1})".format(self[0], self[1])

    def __str__(self):
        """str(self)"""
        return "[{0}, {1}]".format(self[0], self[1])

    def __dealloc__(self):
        del self.thisptr

    @property
    def x(self):
        """First coordinate of IntPoint."""
        return self.thisptr.x()

    @property
    def y(self):
        """Second coordinate of IntPoint."""
        return self.thisptr.y()

    def __add__(cy_IntPoint self, cy_IntPoint o):
        """P + Q"""
        return wrap_IntPoint(deref(self.thisptr)+deref( o.thisptr ))

    def __sub__(cy_IntPoint self, cy_IntPoint o):
        """P - Q"""
        return wrap_IntPoint(deref(self.thisptr)-deref( o.thisptr ))

    def __richcmp__(cy_IntPoint self, cy_IntPoint other, int op):
        if op == 0:
            return deref(self.thisptr) < deref(other.thisptr)
        if op == 1:
            return deref(self.thisptr) <= deref(other.thisptr)
        if op == 2:
            return deref(self.thisptr) == deref(other.thisptr)
        if op == 3:
            return deref(self.thisptr) != deref(other.thisptr)
        if op == 4:
            return deref(self.thisptr) > deref(other.thisptr)
        if op == 5:
            return deref(self.thisptr) >= deref(other.thisptr)

cdef cy_IntPoint wrap_IntPoint(IntPoint p):
    cdef IntPoint * retp = new IntPoint()
    retp[0] = p
    cdef cy_IntPoint r = cy_IntPoint.__new__(cy_IntPoint)
    r.thisptr = retp
    return r


cdef class cy_Line:

    """Class representing line in plane.

    Corresponds to Line class in 2geom. Most members are direct
    calls to Line methods, otherwise C++ call is specified.
    """

    cdef Line* thisptr

    def __cinit__(self, cy_Point cp = None, double x = 0):
        """Create Line from point and angle to x-axis.

        Constructor with no arguments calls Line() in 2geom, otherwise
        Line(Point &, double) is called.
        """
        if cp is None:
            self.thisptr = new Line()
        else:
            self.thisptr = new Line( deref(cp.thisptr), x)

    def __repr__(self):
        """repr(self)."""
        return "Line({0}, {1:3f})".format(repr(self.origin()), self.angle())

    def __str__(self):
        """str(self)"""
        return repr(self)

    def __dealloc__(self):
        del self.thisptr

    @classmethod
    def from_points(cls, cy_Point cp, cy_Point cq):
        """Create Line passing through two points.

        Calls Line(Point &, Point &) in 2geom.
        """
        return wrap_Line( Line( deref(cp.thisptr), deref(cq.thisptr) ) )

    @classmethod
    def from_origin_and_versor(cls, cy_Point o, cy_Point v):
        """Create Line passing through point with specified versor."""
        return wrap_Line( from_origin_and_versor(deref(o.thisptr), deref(v.thisptr)) )

    @classmethod
    def from_normal_distance(cls, cy_Point normal, double dist):
        """Create Line from it's normal and distance from origin."""
        return wrap_Line( from_normal_distance( deref(normal.thisptr), dist ) )

    @classmethod
    def from_LineSegment(cls, cy_LineSegment LS):
        """Create Line from LineSegment.

        Calls Line(LineSegment &) in 2geom.
        """
        return wrap_Line( Line( deref(<LineSegment *> LS.thisptr) ) )

    @classmethod
    def from_Ray(cls, cy_Ray R):
        """Create Line from Ray.

        Calls Line(Ray &) in 2geom.
        """
        return wrap_Line( Line( deref(R.thisptr) ) )

    #maybe implement as properties.

    def origin(self):
        """Return origin of line."""
        return wrap_Point(self.thisptr.origin())

    def versor(self):
        """Return versor of line."""
        return wrap_Point(self.thisptr.versor())

    def angle(self):
        """Return angle between line and x-axis."""
        return self.thisptr.angle()

    def set_origin(self, cy_Point origin):
        """Set origin."""
        self.thisptr.setOrigin( deref(origin.thisptr) )

    def set_versor(self, cy_Point versor):
        """Set versor."""
        self.thisptr.setVersor( deref(versor.thisptr) )

    def set_angle(self, Coord a):
        """Set angle."""
        self.thisptr.setAngle(a)

    def set_points(self, cy_Point cp, cy_Point cq):
        """Set two points line passes through."""
        self.thisptr.setPoints( deref(cp.thisptr), deref(cq.thisptr) )

    def set_coefficients(self, a, b, c):
        """Set coefficients in parametric equation of line."""
        self.thisptr.setCoefficients(a, b, c)

    def is_degenerate(self):
        """Test whether line's versor is zero vector."""
        return self.thisptr.isDegenerate()

    def point_at(self, t):
        """origin + t*versor"""
        return wrap_Point(self.thisptr.pointAt(t))

    def value_at(self, t, Dim2 d):
        """Coordinates of point_at(t)."""
        return self.thisptr.valueAt(t, d)

    def time_at(self, cy_Point cp):
        """Find time value corresponding to point on line."""
        return self.thisptr.timeAt( deref(cp.thisptr) )

    def time_at_projection(self, cy_Point cp):
        """Find time value corresponding to orthogonal projection of point."""
        return self.thisptr.timeAtProjection( deref(cp.thisptr) )

    def nearest_time(self, cy_Point cp):
        """Alias for time_at_projection."""
        return self.thisptr.nearestTime( deref(cp.thisptr) )

    def roots(self, Coord v, Dim2 d):
        """Return time values where self.value_at(t, dim) == v."""
        return wrap_vector_double( self.thisptr.roots(v, d) )

    def reverse(self):
        """Reverse line."""
        return wrap_Line( self.thisptr.reverse() )

    def derivative(self):
        """Take line's derivative."""
        return wrap_Line( self.thisptr.derivative() )

    def normal(self):
        """Return line's normal."""
        return wrap_Point( self.thisptr.normal() )

    def normal_and_dist(self):
        """return tuple containing normal vector and distance from origin.

        Calls normal_and_dist(x) and return it's result and x as a tuple.
        """
        cdef double x = 0
        cdef Point p = self.thisptr.normalAndDist(x)
        return (wrap_Point(p), x)

    def portion(self, Coord f, Coord t):
        """Return Curve corresponding to portion of line."""
        return wrap_Curve_p( self.thisptr.portion(f, t) )

    def ray(self, Coord t):
        """Return Ray continuing from time value t."""
        return wrap_Ray( self.thisptr.ray(t) )

    def segment(self, Coord f, Coord t):
        """Return LineSegment corresponding to portion of line."""
        return wrap_LineSegment( self.thisptr.segment(f, t) )

    def transformed(self, t):
        """Return line transformed by transform."""
        #doing this because transformed(t) takes reference
        cdef Affine at
        if is_transform(t):
            at = get_Affine(t)
            return wrap_Line(self.thisptr.transformed( at ))

    @classmethod
    def distance(cls, cy_Point cp, cy_Line cl):
        """Calculate distance between point and line."""
        return distance( deref(cp.thisptr), deref(cl.thisptr))

    @classmethod
    def are_near(cls, cy_Point cp, cy_Line cl, double eps=EPSILON):
        """Test if point is near line."""
        return are_near( deref(cp.thisptr), deref(cl.thisptr), eps)

    @classmethod
    def are_parallel(cls, cy_Line cl, cy_Line ck, eps=EPSILON):
        """Test if lines are almost parallel."""
        return are_parallel( deref(cl.thisptr), deref(ck.thisptr), eps)

    @classmethod
    def are_same(cls, cy_Line cl, cy_Line ck, double eps=EPSILON):
        """Test if lines represent the same line."""
        return are_same( deref(cl.thisptr), deref(ck.thisptr), eps)

    @classmethod
    def are_orthogonal(cls, cy_Line cl, cy_Line ck, eps=EPSILON):
        """Test two lines for orthogonality."""
        return are_orthogonal( deref(cl.thisptr), deref(ck.thisptr), eps)

    @classmethod
    def are_collinear(cls, cy_Point cp, cy_Point cq, cy_Point cr, eps=EPSILON):
        """Test for collinearity of vectors (cq-cp) and (cr-cp)"""
        return are_collinear( deref(cp.thisptr), deref(cq.thisptr), deref(cr.thisptr), eps)

    @classmethod
    def angle_between(cls, cy_Line cl, cy_Line ck):
        """Calculate angle between two lines"""
        return angle_between( deref(cl.thisptr), deref(ck.thisptr) )

cdef cy_Line wrap_Line(Line p):
    cdef Line * retp = new Line()
    retp[0] = p
    cdef cy_Line r = cy_Line.__new__(cy_Line)
    r.thisptr = retp
    return r

#-- Ray --

cdef class cy_Ray:

    """Ray represents half of line, starting at origin and going to
    infinity.

    Corresponds to Ray class in 2geom. Most members are direct
    calls to Ray methods, otherwise C++ call is specified.
    """

    cdef Ray* thisptr

    def __cinit__(self, cy_Point cp = None, double x = 0):
        """Create Ray from origin and angle with x-axis.

        Empty constructor calls Ray() in 2geom.
        """
        if cp is None:
            self.thisptr = new Ray()
        else:
            self.thisptr = new Ray( deref(cp.thisptr), x)

    def __repr__(self):
        """repr(self)."""
        return "Ray({0}, {1:3f})".format(repr(self.origin()), self.angle())

    def __str__(self):
        """str(self)"""
        return repr(self)

    def __dealloc__(self):
        del self.thisptr

    @classmethod
    def from_points(cls, cy_Point cp, cy_Point cq):
        """Create ray passing through two points, starting at first one."""
        return wrap_Ray( (Ray( deref(cp.thisptr), deref(cq.thisptr) )) )

    def origin(self):
        """Return origin."""
        return wrap_Point(self.thisptr.origin())

    def versor(self):
        """Return versor."""
        return wrap_Point(self.thisptr.versor())

    def angle(self):
        """Return angle between ray and x-axis."""
        return self.thisptr.angle()

    def set_origin(self, cy_Point cp):
        """Set origin."""
        self.thisptr.setOrigin( deref(cp.thisptr) )

    def set_versor(self, cy_Point cp):
        """Set versor."""
        self.thisptr.setVersor( deref(cp.thisptr) )

    def set_angle(self, Coord a):
        """Set angle."""
        self.thisptr.setAngle(a)

    def set_points(self, cy_Point cp, cy_Point cq):
        """Set origin and second point of ray."""
        self.thisptr.setPoints( deref(cp.thisptr), deref(cq.thisptr) )

    def is_degenerate(self):
        """Check for zero versor."""
        return self.thisptr.isDegenerate()

    def point_at(self, t):
        """origin + t * versor"""
        return wrap_Point(self.thisptr.pointAt(t))
    def value_at(self, t, Dim2 d):
        """Access coordinates of point_at(t)."""
        return self.thisptr.valueAt(t, d)

    def nearest_time(self, cy_Point cp):
        """Get time value of nearest point of ray."""
        return self.thisptr.nearestTime( deref(cp.thisptr) )
    def reverse(self):
        """Reverse the ray."""
        return wrap_Ray( self.thisptr.reverse() )

    def roots(self, Coord v, Dim2 d):
        """Return time values for which self.value_at(t, d) == v."""
        return wrap_vector_double( self.thisptr.roots(v, d) )

    def transformed(self, t):
        """Return ray transformed by affine transform."""
        cdef Affine at
        if is_transform(t):
            at = get_Affine(t)
            return wrap_Ray(self.thisptr.transformed( at ))

    def portion(self, Coord f, Coord t):
        """Return Curve corresponding to portion of ray."""
        return wrap_Curve_p( self.thisptr.portion(f, t) )

    def segment(self, Coord f, Coord t):
        """Return LineSegment corresponding to portion of ray."""
        return wrap_LineSegment( self.thisptr.segment(f, t) )

    @classmethod
    def distance(cls, cy_Point cp, cy_Ray cl):
        """Compute distance between point and ray."""
        return distance( deref(cp.thisptr), deref(cl.thisptr))

    @classmethod
    def are_near(cls, cy_Point cp, cy_Ray cl, double eps=EPSILON):
        """Check if distance between point and ray is small."""
        return are_near( deref(cp.thisptr), deref(cl.thisptr), eps)

    @classmethod
    def are_same(cls, cy_Ray cl, cy_Ray ck, double eps=EPSILON):
        """Check if two ray are same."""
        return are_same( deref(cl.thisptr), deref(ck.thisptr), eps)

    @classmethod
    def angle_between(cls, cy_Ray cl, cy_Ray ck, bint cw=True):
        """Compute angle between two rays.

        Can specify direction using parameter cw.
        """
        return angle_between( deref(cl.thisptr), deref(ck.thisptr), cw)

    @classmethod
    def make_angle_bisector_ray(cls, cy_Ray cl, cy_Ray ck):
        """Make ray bisecting smaller angle formed by two rays."""
        return wrap_Ray( make_angle_bisector_ray(deref(cl.thisptr), deref(ck.thisptr) ))

cdef cy_Ray wrap_Ray(Ray p):
    cdef Ray * retp = new Ray()
    retp[0] = p
    cdef cy_Ray r = cy_Ray.__new__(cy_Ray)
    r.thisptr = retp
    return r



from _common_decl cimport *

from libcpp.vector cimport vector
from libcpp.pair cimport pair

from _cy_rectangle cimport Interval, OptInterval, Rect, OptRect
from _cy_affine cimport Affine
from _cy_curves cimport Curve, cy_Curve, wrap_Curve_p
from _cy_curves cimport SBasis, cy_SBasis

from _cy_primitives cimport Point, cy_Point, wrap_Point


cdef extern from "2geom/d2.h" namespace "Geom":
    cdef cppclass D2[T]:
        D2()
        D2(T &, T &)
        T& operator[](unsigned i)  

#~ ctypedef int BaseIteratorConst "BaseIterator<ConstIterator, Path const>"

cdef extern from "2geom/path.h" namespace "Geom::PathInternal":
    cdef cppclass BaseIterator[C, P]:
        Curve & c_item "operator*" ()
        C & c_next "operator++" ()
        C & c_next "operator++" (int)
        bint operator==(BaseIterator[C, P])


    cdef cppclass ConstIterator:
        ConstIterator()
    cdef cppclass Iterator:
        Iterator()
        ConstIterator & operator()        
        
    cdef cppclass BaseIteratorConst "Geom::PathInternal::BaseIterator<Geom::PathInternal::ConstIterator, Path const>":
#~         Curve & c_item "operator*" ()
        ConstIterator & c_next "operator++" ()
        ConstIterator & c_next "operator++" (int)
#~         bint operator==(BaseIterator[C, P])
        
 
cdef extern from "2geom/path.h" namespace "Geom::Path":
    cdef enum Stitching:
        c_NO_STITCHING "Path::NO_STITCHING" = 0,
        c_STITCH_DISCONTINUOUS "Path::STITCH_DISCONTINUOUS" 

cdef extern from "2geom/path.h" namespace "Geom":
    cdef cppclass Path:
        Path(Path &)
        Path(Point)
        Path(ConstIterator &, ConstIterator &, bint)
        void swap(Path &)
        Curve & operator[](unsigned int)
#~         Curve & at_index(unsigned int)
#~         ::boost::shared_ptr<Geom::Curve const> get_ref_at_index(unsigned int)
        Curve & front()
        Curve & back()
        Curve & back_open()
        Curve & back_closed()
        Curve & back_default()
        ConstIterator begin_const "begin" ()
        ConstIterator end()
        Iterator begin()
#~         Iterator end()
        ConstIterator end_open()
        ConstIterator end_closed()
        ConstIterator end_default()
        size_t size_open()
        size_t size_closed()
        size_t size_default()
        size_t size()
        size_t max_size()
        bint empty()
        bint closed()
        void close(bint)
        OptRect boundsFast()
        OptRect boundsExact()
#~         Piecewise<Geom::D2<Geom::SBasis> > toPwSb()
        bint operator==(Path &)
        bint operator!=(Path &)
        Path operator*(Affine &)
#~         Path & operator*=(Affine &)
        Point pointAt(double)
        double valueAt(double, Dim2)
        Point operator()(double) except +
        vector[double] roots(double, Dim2)
        vector[double] allNearestTimes(Point &, double, double)
        vector[double] allNearestTimes(Point &)
        vector[double] nearestTimePerCurve(Point &)
        double nearestTime(Point &, double, double, double *)
        double nearestTime(Point &, double *)
        void appendPortionTo(Path &, double, double)
        Path portion(double, double)
        Path portion(Interval)
        Path reversed()
        void insert(Iterator &, Curve &, Stitching) except +
        void insert(Iterator &, ConstIterator &, ConstIterator &, Stitching)
        void clear()
        void erase(Iterator &, Stitching)
        void erase(Iterator &, Iterator &, Stitching)
        void erase_last()
        void replace(Iterator &, Curve &, Stitching)
        void replace(Iterator &, Iterator &, Curve &, Stitching)
        void replace(Iterator &, ConstIterator &, ConstIterator &, Stitching)
        void replace(Iterator &, Iterator &, ConstIterator &, ConstIterator &, Stitching)
        void start(Point)
        Point initialPoint()
        Point finalPoint()
        void setInitial(Point &)
        void setFinal(Point &)
        void append(Curve &, Stitching)
        void append(D2[SBasis] &, Stitching)
        void append(Path &, Stitching)
        void stitchTo(Point &)

cdef class cy_Path:
#~     NO_STITCHING = c_NO_STITCHING
#~     STITCH_DISCONTINUOUS = c_STITCH_DISCONTINUOUS
    cdef Path* thisptr
    cdef ConstIterator _const_iterator_at_index(self, int i)
    cdef Iterator _iterator_at_index(self, int i)

cdef cy_Path wrap_Path(Path p)



#!/usr/bin/python

import py2geom
import toyframework
import random,gtk
from py2geom_glue import *

def cairo_pw(cr, p):
    for i in range(p.size()):
        a,b = p.cuts[i], p.cuts[i+1]
        bez = sbasis_to_bezier(p[i], 0)
        cr.move_to(a, bez[0])
        cr.curve_to(lerp(a, b, 1./3), bez[1],
                    lerp(a,b, 2./3), bez[2],
                    b, bez[3])

def cairo_horiz(cr, y, ps):
    for p in ps:
        cr.move_to(p, y);
        cr.rel_line_to(0, 10)

def cairo_vert(cr, x, ps):
    for p in ps:
        cr.move_to(x, p)
        cr.rel_line_to(10, 0)

def l2s(l):
    sb = py2geom.SBasis(l)
    return sb

def constant(l):
    pws = py2geom.PiecewiseSBasis()
    
    pws.push_cut(0)
    pws.push_seg(l2s(py2geom.Linear(l,l)) )
    pws.push_cut(1000)
    return pws

X = l2s(py2geom.Linear(0, 1))
OmX = l2s(py2geom.Linear(1, 0))
def bezier_to_sbasis(handles, order):
    if(order == 0):
        return l2s(py2geom.Linear(handles[0]))
    elif(order == 1):
        return l2s(py2geom.Linear(handles[0], handles[1]))
    else:
        return (py2geom.multiply(OmX, bezier_to_sbasis(handles[:-1], order-1)) +
                py2geom.multiply(X, bezier_to_sbasis(handles[1:], order-1)))


def bez_to_sbasis(handles, order):
    return bezier_to_sbasis([h[1] for h in handles], order)

class PWSBHandle(toyframework.Handle):
    def __init__(self, cs, segs):
        self.handles_per_curve = cs*segs
        self.curve_size = cs
        self.segs = segs
        self.pts = []
    def append(self, x, y):
        self.pts.append(py2geom.Point(x,y))
    def value(self, y_0=0):
        pws = py2geom.PiecewiseSBasis()
        for i in range(0, self.handles_per_curve, self.curve_size):
	    pws.push_cut(self.pts[i][0]);
	    for j in range(i, i + self.curve_size):
		self.pts[j] = py2geom.Point(self.pts[j][0], self.pts[j][1] - y_0)
            hnd = self.pts[i:i+self.curve_size]
            pws.push_seg( bez_to_sbasis(hnd, self.curve_size-1));
	    for j in range(i, i + self.curve_size):
		self.pts[j] = py2geom.Point(self.pts[j][0], self.pts[j][1] + y_0);

	pws.push_cut(self.pts[self.handles_per_curve - 1][0]);
	assert(pws.invariants());
	return pws

    def draw(self, cr, annotes):
        for p in self.pts:
            toyframework.draw_circ(cr, p)

    def hit(self, mouse):
        for i,p in enumerate(self.pts):
            if(py2geom.distance(py2geom.Point(*mouse), p) < 5):
                return i
        return None

    def move_to(self, hit, om, m):
        om = py2geom.Point(*om)
        m = py2geom.Point(*m)
        if hit != None:
            i,hand = hit
            self.pts[hand] = m
            for i in range(self.curve_size, self.handles_per_curve, self.curve_size):
                self.pts[i-1] = py2geom.Point(self.pts[i][0],self.pts[i-1][1])

            for i in range(0, self.handles_per_curve, self.curve_size):
                for j in range(1, (self.curve_size-1)):
                    t = float(j)/(self.curve_size-1)
                    x = lerp(self.pts[i][0], self.pts[i+self.curve_size-1][0],t)
                    self.pts[i+j] = py2geom.Point(x, self.pts[i+j][1])



class PwToy(toyframework.Toy):
    def __init__(self):
        toyframework.Toy.__init__(self)
        self.segs = 2
        self.handles_per_curve = 4 * self.segs
        self.curves = 2
        self.pwsbh = []
        self.interval_test = []
        for a in range(self.curves):
            self.pwsbh.append(PWSBHandle(4, self.curves))
            self.handles.append(self.pwsbh[a])
            for i in range(self.handles_per_curve):
                t = 150 + 300*i/(4*self.segs)
                self.pwsbh[a].append(t, random.uniform(0,1) * 150 + 150 - 50 * a)
        self.interval_test.append(toyframework.PointHandle(150, 400))
        self.interval_test.append(toyframework.PointHandle(300, 400))
        self.handles.append(self.interval_test[0])
        self.handles.append(self.interval_test[1])
        self.func = "pws[0] + pws[1]"
    def gtk_ready(self):
        import gtk
        self.sb_entry = gtk.Entry()
        self.sb_entry.connect("changed", self.sb_changed)
        toyframework.get_vbox().add(self.sb_entry)
        toyframework.get_vbox().show_all()
        self.sb_entry.set_text(self.func)
    def sb_changed(self, sb):
        self.func = sb.get_text()
        self.redraw()
    def draw(self, cr, pos, save):
        cr.set_source_rgba (0., 0., 0., 1)
        cr.set_line_width (1)
        
        pws = [self.pwsbh[i].value() for i in range(self.curves)]
        for p in pws:
            cairo_pw(cr, p)
        cr.stroke()
        
        d = locals().copy()
        for i in dir(py2geom):
            d[i] = py2geom.__dict__[i]
        d['l2s'] = l2s
        d['constant'] = constant
        pw_out = eval(self.func, d)

        bs = py2geom.bounds_local(pw_out, py2geom.OptInterval(
                          py2geom.Interval(self.interval_test[0].pos[0], 
                                           self.interval_test[1].pos[0])));
        if not bs.isEmpty():
            bs = bs.toInterval()
            for ph in self.interval_test:
                ph.pos= py2geom.Point(ph.pos[0], bs.middle())
            cr.save()
            cr.set_source_rgba (.0, 0.25, 0.5, 1.)
            cr.rectangle(self.interval_test[0].pos[0], bs.min(),
                         self.interval_test[1].pos[0]-self.interval_test[0].pos[0], bs.extent())
            cr.stroke()
        bs = py2geom.bounds_exact(pw_out);
        cr.set_source_rgba (0.25, 0.25, .5, 1.);
        if not bs.isEmpty():
            bs = bs.toInterval()
            cairo_horiz(cr, bs.middle(), pw_out.cuts);
            cr.stroke()
        cr.restore()

        cr.set_source_rgba (0., 0., .5, 1.);
        cairo_pw(cr, pw_out)
        cr.stroke()


        self.notify = str(bs)
        toyframework.Toy.draw(self, cr, pos, save)
        
t = PwToy()
import sys

toyframework.init(sys.argv, t, 500, 500)



from cython.operator cimport dereference as deref

from numbers import Number


cdef class cy_Affine:

    """Class representing affine transform in 2D plane.

    Corresponds to Affine class in 2geom.
    """

    def __cinit__(self, c0=None,
                        Coord c1=0,
                        Coord c2=0,
                        Coord c3=1,
                        Coord c4=0,
                        Coord c5=0):
        """Create Affine instance from either transform or from coefficients."""
        if c0 is None:
            self.thisptr = new Affine()
        elif is_transform(c0):
            self.thisptr = new Affine( get_Affine(c0) )
        else:
            self.thisptr = new Affine(<Coord> float(c0) ,c1 ,c2 ,c3 ,c4 ,c5)

    def __str__(self):
        """str(self)"""
        return "Affine({}, {}, {}, {}, {}, {})".format( self[0],
                                                        self[1],
                                                        self[2],
                                                        self[3],
                                                        self[4],
                                                        self[5],
                                                        )
    def __repr__(self):
        """repr(self)"""
        return str(self)

    def __dealloc__(self):
        del self.thisptr

    def __getitem__(self, int i):
        """Get coefficients."""
        if i >= 6:
            raise IndexError("Affine has only 6 coefficients.")
        return deref(self.thisptr) [i]

    def __mul__(cy_Affine self, other):
        """Compose with another transformation."""
        if isinstance(other, cy_Affine):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_Affine> other).thisptr ) )
        elif isinstance(other, cy_Translate):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_Translate> other).thisptr ) )
        elif isinstance(other, cy_Scale):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_Scale> other).thisptr ) )
        elif isinstance(other, cy_Rotate):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_Rotate> other).thisptr ) )
        elif isinstance(other, cy_HShear):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_HShear> other).thisptr ) )
        elif isinstance(other, cy_VShear):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_VShear> other).thisptr ) )
        elif isinstance(other, cy_Zoom):
            return wrap_Affine( deref(self.thisptr) * deref( (<cy_Zoom> other).thisptr ) )

    def __pow__(cy_Affine self, int n, z):
        """Compose with self n times."""
        return wrap_Affine(pow( deref(self.thisptr), n ))

    def __richcmp__(cy_Affine self, cy_Affine other, int op):
        if op == 2:
            return deref(self.thisptr) == deref(other.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(other.thisptr)

    def x_axis(self):
        """Transformation of unit x vector without translation."""
        return wrap_Point(self.thisptr.xAxis())

    def y_axis(self):
        """Transformation of unit y vector without translation."""
        return wrap_Point(self.thisptr.yAxis())

    def translation(self):
        """Translation of transformation."""
        return wrap_Point(self.thisptr.translation())

    def expansion_X(self):
        """Expansion of unit x vector."""
        return self.thisptr.expansionX()

    def expansion_Y(self):
        """Expansion of unit y vector."""
        return self.thisptr.expansionY()

    def expansion(self):
        """Point( self.expansion_X(), self.expansion_Y() )"""
        return wrap_Point(self.thisptr.expansion())

    def set_X_axis(self, cy_Point vec):
        """Set transformation of x unit vector without translation."""
        self.thisptr.setXAxis(deref( vec.thisptr ))

    def set_Y_axis(self, cy_Point vec):
        """Set transformation of y unit vector without translation."""
        self.thisptr.setYAxis(deref( vec.thisptr ))

    def set_translation(self, cy_Point loc):
        """Set translation of origin."""
        self.thisptr.setTranslation(deref( loc.thisptr ))

    def set_expansion_X(self, Coord val):
        """Set expansion of x unit vector."""
        self.thisptr.setExpansionX(val)

    def set_expansion_Y(self, Coord val):
        """Set expansion of y unit vector."""
        self.thisptr.setExpansionY(val)

    def set_identity(self):
        """Set self to identity transformation."""
        self.thisptr.setIdentity()

    def is_identity(self, Coord eps=EPSILON):
        """Return true if self is close to identity transform.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isIdentity(eps)

    def is_translation(self, Coord eps=EPSILON):
        """Return true if self is close to transformation.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isTranslation(eps)

    def is_scale(self, Coord eps=EPSILON):
        """Return true if self is close to scale.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isScale(eps)

    def is_uniform_scale(self, Coord eps=EPSILON):
        """Return true if self is close to uniform scale.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isUniformScale(eps)

    def is_rotation(self, Coord eps=EPSILON):
        """Return true if self is close to rotation.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isRotation(eps)

    def is_HShear(self, Coord eps=EPSILON):
        """Return true if self is close to horizontal shear.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isHShear(eps)

    def is_VShear(self, Coord eps=EPSILON):
        """Return true if self is close to vertical shear.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isVShear(eps)

    def is_nonzero_translation(self, Coord eps=EPSILON):
        """Return true if self is close to translation and is identity.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isNonzeroTranslation(eps)

    def is_nonzero_scale(self, Coord eps=EPSILON):
        """Return true if self is close to scale and is identity.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isNonzeroScale(eps)

    def is_nonzero_uniform_scale(self, Coord eps=EPSILON):
        """Return true if self is close to scale and is identity.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isNonzeroUniformScale(eps)

    def is_nonzero_rotation(self, Coord eps=EPSILON):
        """Return true if self is close to rotation and is identity.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isNonzeroRotation(eps)

    def is_nonzero_HShear(self, Coord eps=EPSILON):
        """Return true if self is close to horizontal shear and is not identity.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isNonzeroHShear(eps)

    def is_nonzero_VShear(self, Coord eps=EPSILON):
        """Return true if self is close to vertical shear and is not identity.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isNonzeroVShear(eps)

    def is_zoom(self, Coord eps=EPSILON):
        """Return true if self is close to zoom.

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.isZoom(eps)

    def preserves_area(self, Coord eps=EPSILON):
        """Return true if areas are preserved after transformation

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.preservesArea(eps)

    def preserves_angles(self, Coord eps=EPSILON):
        """Return true if angles are preserved after transformation

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.preservesAngles(eps)

    def preserves_distances(self, Coord eps=EPSILON):
        """Return true if distances are preserved after transformation

        Use second argument eps to specify tolerance.
        """
        return self.thisptr.preservesDistances(eps)

    def flips(self):
        """Return true if transformation flips - it has negative scaling."""
        return self.thisptr.flips()

    def is_singular(self, Coord eps=EPSILON):
        """Check whether transformation matrix is singular."""
        return self.thisptr.isSingular(eps)

    def without_translation(self):
        """Return transformation without translation part."""
        return wrap_Affine(self.thisptr.withoutTranslation())

    def inverse(self):
        """Return inverse transformation."""
        return wrap_Affine(self.thisptr.inverse())

    def det(self):
        """Return determinant of transformation matrix."""
        return self.thisptr.det()

    def descrim2(self):
        """Return absolute value of self.det()"""
        return self.thisptr.descrim2()

    def descrim(self):
        """Return square root of self.descrim2()"""
        return self.thisptr.descrim()

    @classmethod
    def identity(self):
        """Create identity transformation."""
        return wrap_Affine(a_identity())

    @classmethod
    def are_near(cls, A, B, Coord eps=EPSILON):
        """Test if two transforms are near."""
        if is_transform(A) & is_transform(B):
            return are_near(get_Affine(A), get_Affine(B), eps)

    @classmethod
    def reflection(cls, cy_Point vector, cy_Point origin):
        """Create transformation reflecting along line specified by vector and origin."""
        return wrap_Affine( reflection( deref(vector.thisptr), deref(origin.thisptr) ) )

cdef cy_Affine wrap_Affine(Affine p):
    cdef Affine * retp = new Affine()
    retp[0] = p
    cdef cy_Affine r = cy_Affine.__new__(cy_Affine)
    r.thisptr = retp
    return r

cdef Affine get_Affine(t):
    if isinstance(t, cy_Affine ):
        return deref( (<cy_Affine> t).thisptr )
    elif isinstance(t, cy_Translate):
        return <Affine> deref( (<cy_Translate> t).thisptr )
    elif isinstance(t, cy_Scale):
        return <Affine> deref( (<cy_Scale> t).thisptr )
    elif isinstance(t, cy_Rotate):
        return <Affine> deref( (<cy_Rotate> t).thisptr )
    elif isinstance(t, cy_HShear):
        return <Affine> deref( (<cy_HShear> t).thisptr )
    elif isinstance(t, cy_VShear):
        return <Affine> deref( (<cy_VShear> t).thisptr )
    elif isinstance(t, cy_Zoom):
        return <Affine> deref( (<cy_Zoom> t).thisptr )

cdef bint is_transform(t):
    return any([isinstance(t, cy_Affine),
                isinstance(t, cy_Translate),
                isinstance(t, cy_Scale),
                isinstance(t, cy_Rotate),
                isinstance(t, cy_HShear),
                isinstance(t, cy_VShear),
                isinstance(t, cy_Zoom)
                ])


cdef class cy_Translate:

    """Translation in 2D plane

    Corresponds to Translate class in 2geom.
    """

    def __cinit__(self, *args):
        """Create Translate instance form point or it's two coordinates."""
        if len(args) == 0:
            self.thisptr = new Translate()
        elif len(args) == 1:
            self.thisptr = new Translate( deref( (<cy_Point> args[0]).thisptr ) )
        elif len(args) == 2:
            self.thisptr = new Translate(float(args[0]), float(args[1]))

    def __dealloc__(self):
        del self.thisptr


    def __getitem__(self, Dim2 dim):
        """Get components of displacement vector."""
        return deref( self.thisptr ) [dim]

    def __mul__(cy_Translate self, o):
        """Compose with another transformation."""
        if isinstance(o, cy_Translate):
            return wrap_Translate(deref( self.thisptr ) * deref( (<cy_Translate>o).thisptr ))
        elif is_transform(o):
            return wrap_Affine(deref(self.thisptr) * get_Affine(o))

    def __pow__(cy_Translate self, int n, z):
        """Compose with self n times."""
        return wrap_Translate(pow( deref(self.thisptr), n ))

    def vector(self):
        """Get displacement vector."""
        return wrap_Point(self.thisptr.vector())

    def inverse(self):
        """Return inverse transformation."""
        return wrap_Translate(self.thisptr.inverse())

    @classmethod
    def identity(self):
        """Create identity translation."""
        return wrap_Translate(t_identity())

    def __richcmp__(cy_Translate self, cy_Translate t, op):
        if op == 2:
            return deref(self.thisptr) == deref(t.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(t.thisptr)

cdef cy_Translate wrap_Translate(Translate p):
    cdef Translate * retp = new Translate()
    retp[0] = p
    cdef cy_Translate r = cy_Translate.__new__(cy_Translate)
    r.thisptr = retp
    return r


cdef class cy_Scale:

    """Scale in 2D plane.

    Corresponds to Scale in 2geom.
    """

    def __cinit__(self, *args):
        """Create scale from number, point or it's two coordinates.

        One number creates uniform scale, point or two numbers create
        scale with different x and y scale factor.
        """
        if len(args) == 0:
            self.thisptr = new Scale()
        elif len(args) == 1:
            if isinstance(args[0], Number):
                self.thisptr = new Scale(<Coord> float(args[0]))
            elif isinstance(args[0], cy_Point):
                self.thisptr = new Scale( deref( (<cy_Point> args[0]).thisptr ) )
        elif len(args) == 2:
            self.thisptr = new Scale(float(args[0]), float(args[1]))

    def __dealloc__(self):
        del self.thisptr

    def __getitem__(self, Dim2 d):
        """Get scale factors for each axis."""
        return deref( self.thisptr ) [d]

    def __mul__(cy_Scale self, o):
        """Compose with another transformation."""
        if isinstance(o, cy_Scale):
            return wrap_Scale(deref( self.thisptr ) * deref( (<cy_Scale>o).thisptr ))
        elif is_transform(o):
            return wrap_Affine(deref(self.thisptr) * get_Affine(o))

    def __pow__(cy_Scale self, int n, z):
        """Compose with self n times."""
        return wrap_Scale(pow( deref(self.thisptr), n ))

    def vector(self):
        """Get both scale factors as a point."""
        return wrap_Point(self.thisptr.vector())

    def inverse(self):
        """Return inverse transformation."""
        return wrap_Scale(self.thisptr.inverse())

    @classmethod
    def identity(self):
        """Create identity scale."""
        return wrap_Scale(s_identity())

    def __richcmp__(cy_Scale self, cy_Scale s, op):
        if op == 2:
            return deref(self.thisptr) == deref(s.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(s.thisptr)

cdef cy_Scale wrap_Scale(Scale p):
    cdef Scale * retp = new Scale()
    retp[0] = p
    cdef cy_Scale r = cy_Scale.__new__(cy_Scale)
    r.thisptr = retp
    return r


cdef class cy_Rotate:

    """Rotation in 2D plane.

    Corresponds to Rotate in 2geom.
    """

    def __cinit__(self, *args):
        """Create new Rotate instance, specifying angle.

        Use one number to set the angle, or point/its two coordinates,
        using point's angle with x-axis as a rotation angle.
        """
        if len(args) == 0:
            self.thisptr = new Rotate()
        elif len(args) == 1:
            if isinstance(args[0], Number):
                self.thisptr = new Rotate(<Coord> float(args[0]))
            elif isinstance(args[0], cy_Point):
                self.thisptr = new Rotate( deref( (<cy_Point> args[0]).thisptr ) )
        elif len(args) == 2:
            self.thisptr = new Rotate(float(args[0]), float(args[1]))

    def __dealloc__(self):
        del self.thisptr

    def vector(self):
        """Return Point(1, 0)*self."""
        return wrap_Point(self.thisptr.vector())

    def __getitem__(self, Dim2 dim):
        """Get components of self.vector()"""
        return deref( self.thisptr ) [dim]

    def __mul__(cy_Rotate self, o):
        """Compose with another transformation."""
        if isinstance(o, cy_Rotate):
            return wrap_Rotate(deref( self.thisptr ) * deref( (<cy_Rotate>o).thisptr ))
        elif is_transform(o):
            return wrap_Affine(deref(self.thisptr) * get_Affine(o))

    def __pow__(cy_Rotate self, int n, z):
        """Compose with self n times."""
        return wrap_Rotate(pow( deref(self.thisptr), n ))

    def inverse(self):
        """Return inverse transformation."""
        return wrap_Rotate(self.thisptr.inverse())

    @classmethod
    def identity(cls):
        """Create identity rotation."""
        return wrap_Rotate(r_identity())

    @classmethod
    def from_degrees(cls, Coord deg):
        """Create rotation from angle in degrees."""
        return wrap_Rotate(from_degrees(deg))

    def __richcmp__(cy_Rotate self, cy_Rotate r, op):
        if op == 2:
            return deref(self.thisptr) == deref(r.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(r.thisptr)

cdef cy_Rotate wrap_Rotate(Rotate p):
    cdef Rotate * retp = new Rotate()
    retp[0] = p
    cdef cy_Rotate r = cy_Rotate.__new__(cy_Rotate)
    r.thisptr = retp
    return r


cdef class cy_VShear:

    """Vertical shear in 2D plane

    Corresponds to VShear in 2geom.
    """

    def __cinit__(self, Coord h):
        """Create VShear instance form shearing factor."""
        self.thisptr = new VShear(h)

    def __dealloc__(self):
        del self.thisptr

    def factor(self):
        """Get shearing factor."""
        return self.thisptr.factor()

    def set_factor(self, Coord nf):
        """Set shearing factor."""
        self.thisptr.setFactor(nf)

    def __mul__(cy_VShear self, o):
        """Compose with another transformation."""
        if isinstance(o, cy_VShear):
            return wrap_VShear(deref( self.thisptr ) * deref( (<cy_VShear>o).thisptr ))
        elif is_transform(o):
            return wrap_Affine(deref(self.thisptr) * get_Affine(o))

    def __pow__(cy_VShear self, int n, z):
        """Compose with self n times."""
        return wrap_VShear(pow( deref(self.thisptr), n ))

    def inverse(self):
        """Return inverse transformation."""
        return wrap_VShear(self.thisptr.inverse())

    @classmethod
    def identity(cls):
        """Create identity VShear."""
        return wrap_VShear( vs_identity() )

    def __richcmp__(cy_VShear self, cy_VShear hs, op):
        if op == 2:
            return deref(self.thisptr) == deref(hs.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(hs.thisptr)



cdef cy_VShear wrap_VShear(VShear p):
    cdef VShear * retp = new VShear(0)
    retp[0] = p
    cdef cy_VShear r = cy_VShear.__new__(cy_VShear, 0)
    r.thisptr = retp
    return r

cdef class cy_HShear:

    """Horizontal shear in 2D plane

    Corresponds to HShear in 2geom.
    """

    def __cinit__(self, Coord h):
        """Create HShear instance form shearing factor."""
        self.thisptr = new HShear(h)

    def __dealloc__(self):
        del self.thisptr

    def factor(self):
        """Get shearing factor."""
        return self.thisptr.factor()

    def set_factor(self, Coord nf):
        """Set shearing factor."""
        self.thisptr.setFactor(nf)

    def __mul__(cy_HShear self, o):
        """Compose with another transformation."""
        if isinstance(o, cy_HShear):
            return wrap_HShear(deref( self.thisptr ) * deref( (<cy_HShear>o).thisptr ))
        elif is_transform(o):
            return wrap_Affine(deref(self.thisptr) * get_Affine(o))

    def __pow__(cy_HShear self, int n, z):
        """Compose with self n times."""
        return wrap_HShear(pow( deref(self.thisptr), n ))

    def inverse(self):
        """Return inverse transformation."""
        return wrap_HShear(self.thisptr.inverse())

    @classmethod
    def identity(cls):
        """Create identity HShear."""
        return wrap_HShear( hs_identity() )

    def __richcmp__(cy_HShear self, cy_HShear hs, op):
        if op == 2:
            return deref(self.thisptr) == deref(hs.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(hs.thisptr)

cdef cy_HShear wrap_HShear(HShear p):
    cdef HShear * retp = new HShear(0)
    retp[0] = p
    cdef cy_HShear r = cy_HShear.__new__(cy_HShear, 0)
    r.thisptr = retp
    return r


cdef class cy_Zoom:

    """Zoom in 2D plane, consisting of uniform scale and translation.

    Corresponds to Zoom in 2geom.
    """

    def __cinit__(self, Coord scale=1, cy_Translate translate=cy_Translate()):
        """Create Zoom from scale factor and translation"""
        self.thisptr = new Zoom( scale, deref( translate.thisptr ) )

    def __dealloc__(self):
        del self.thisptr

    def __mul__(cy_Zoom self, cy_Zoom z):
        """Compose with another transformation."""
        return wrap_Zoom( deref(self.thisptr) * deref( z.thisptr ))

    def __pow__(cy_Zoom self, int n, z):
        """Compose with self n times."""
        return wrap_Zoom(pow( deref(self.thisptr), n ))

    def __richcmp__(cy_Zoom self, cy_Zoom z, op):
        if op == 2:
            return deref(self.thisptr) == deref(z.thisptr)
        elif op == 3:
            return deref(self.thisptr) != deref(z.thisptr)

    def scale(self):
        """Get scale factor."""
        return self.thisptr.scale()

    def set_scale(self, Coord s):
        """Set scale factor."""
        self.thisptr.setScale(s)

    def translation(self):
        """Get translation as a point."""
        return wrap_Point(self.thisptr.translation())

    def set_translation(self, cy_Point p):
        """Set translation."""
        self.thisptr.setTranslation(deref( p.thisptr ))

    def inverse(self):
        """Return inverse transformation."""
        return wrap_Zoom(self.thisptr.inverse())

    @classmethod
    def identity(cls):
        """Create identity zoom."""
        return wrap_Zoom(z_identity())

    @classmethod
    def map_rect(self, cy_Rect old_r, cy_Rect new_r):
        """Create zooming used to go from old rectangle to new."""
        return wrap_Zoom(map_rect(deref( old_r.thisptr ) ,deref( new_r.thisptr )))

cdef cy_Zoom wrap_Zoom(Zoom p):
    cdef Zoom * retp = new Zoom(0)
    retp[0] = p
    cdef cy_Zoom r = cy_Zoom.__new__(cy_Zoom, 0)
    r.thisptr = retp
    return r


cdef class cy_Eigen:

    """Class computing eigenvalues and eigenvectors of 2x2 matrix.

    Corresponds to Eigen class in 2geom.
    """

    cdef Eigen* thisptr

    def __cinit__(self, a):
        """Create Eigen form 2D transform or 2x2 list - matrix."""
        cdef Affine at
        cdef double m[2][2]
        if is_transform(a):
            at = get_Affine(a)
            self.thisptr = new Eigen(at)
        else:
            for i in range(2):
                for j in range(2):
                    m[i][j] = a[i][j]
            self.thisptr = new Eigen(m)

    def __dealloc__(self):
        del self.thisptr

    @property
    def vectors(self):
        """Eigenvectors of matrix."""
        return (wrap_Point(self.thisptr.vectors[0]), wrap_Point(self.thisptr.vectors[1]))

    @property
    def values(self):
        """Eigenvalues of matrix."""
        return (self.thisptr.values[0], self.thisptr.values[1])



from cython.operator cimport dereference as deref
from numbers import Number
from _cy_rectangle cimport cy_OptInterval, wrap_OptInterval, wrap_Rect, OptRect, wrap_OptRect
from _cy_rectangle cimport cy_Interval, wrap_Interval

from _cy_affine cimport cy_Affine, wrap_Affine, get_Affine, is_transform

from _cy_curves cimport is_Curve, get_Curve_p


cdef class cy_Iterator:
    cdef Iterator* thisptr
    cdef ConstIterator* startptr
    cdef ConstIterator* endptr

    def __cinit__(self):
        self.thisptr = new Iterator()

    def __dealloc__(self):
        del self.thisptr

    def __iter__(self):
        return self

    def __next__(self):
        if deref(<BaseIterator[ConstIterator, Path] *>  self.endptr) == deref(<BaseIterator[ConstIterator, Path] *> self.thisptr):
            raise StopIteration
        cdef Curve * r = <Curve *> &(<BaseIterator[Iterator, Path] * > self.thisptr).c_item()
        (<BaseIterator[Iterator, Path] *> self.thisptr).c_next()
        return wrap_Curve_p ( r )




cdef cy_Iterator wrap_Iterator(Iterator i, ConstIterator starti, ConstIterator endi):
    cdef Iterator * retp = new Iterator()
    retp[0] = i
    cdef cy_Iterator r = cy_Iterator.__new__(cy_Iterator)
    r.thisptr = retp

    cdef ConstIterator * endp = new ConstIterator()
    endp[0] = endi
    r.endptr = endp

    cdef ConstIterator * startp = new ConstIterator()
    startp[0] = starti
    r.startptr = startp

    return r

cdef class cy_Path:

    """Path is a ordered sequence of curves.

    You can iterate this class, accessing curves one at time, or access
    them using indices.

    Two constants, NO_STITCHING and STITCH_DISCONTINUOUS are members of
    Path namespace, and are used to specify type of stitching, if
    necessary.

    Path is either open or closed, but in both cases carries closing
    segment, connecting last and first point.

    Corresponds to Path class in 2geom.
    """

    NO_STITCHING = c_NO_STITCHING
    STITCH_DISCONTINUOUS = c_STITCH_DISCONTINUOUS

    def __cinit__(self, cy_Point p=cy_Point()):
        """Create Path containing only one point."""
        self.thisptr = new Path(deref( p.thisptr ))

    @classmethod
    def fromList(cls, l, Stitching stitching=NO_STITCHING, closed=False):
        """Create path from list of curves.

        Specify stithing and closed flag in additional arguments.
        """
        p = cy_Path()
        for curve in l:
            p.append_curve(curve, stitching)
        p.close(closed)
        return p

    @classmethod
    def fromPath(cls, cy_Path p, fr=-1, to = 1, bint closed=False):
        """Create path copying another's path curves.

        Either copy all curves, or ones from
        fr - index of first curve copied
        to
        to -  index of first curve not copied.

        Also takes closed flag.
        """
        if fr == -1:
            return wrap_Path( Path(p.thisptr.begin_const(), p.thisptr.end_default(), closed) )
        else:
            return wrap_Path( Path(p._const_iterator_at_index(fr), p._const_iterator_at_index(to), closed) )

    def __dealloc__(self):
        del self.thisptr

    def __getitem__(self, unsigned int i):
        """Get curve with index i."""
        cdef Curve * r = <Curve *> & deref(self.thisptr)[i]
        return wrap_Curve_p(r)

    def __call__(self, double t):
        """Evaluate path at time t.

        Note: t can be greater than 1 here, it can go to self.size()
        """
        return wrap_Point(deref( self.thisptr ) (t) )

    def __richcmp__(cy_Path self, cy_Path other, int op):
        if op == 2:
            return deref( self.thisptr ) == deref( other.thisptr )
        elif op == 3:
            return deref( self.thisptr ) != deref( other.thisptr )

    def __mul__(cy_Path self, m):
        """Transform path with a transform."""
        cdef Affine at
        if is_transform(m):
            at = get_Affine(m)
            return wrap_Path( deref(self.thisptr) * at )

    #This is not the fastest way, but it's pretty nice from python's perspective
    #Anyway, I would expect that performance hit is minimal, since i is generally really small
    cdef ConstIterator _const_iterator_at_index(self, int i):
        cdef ConstIterator ci = self.thisptr.begin_const()
        cdef ConstIterator * cip = &ci
        for ii in range(i):
            (<BaseIteratorConst *> cip).c_next()
        return ci

    cdef Iterator _iterator_at_index(self, int i):
        cdef Iterator ci = self.thisptr.begin()
        cdef Iterator * cip = &ci
        for ii in range(i):
            (<BaseIterator[Iterator, Path] *> cip).c_next()
        return ci

    def swap(self, cy_Path other):
        """Swap curves with another path."""
        self.thisptr.swap(deref( other.thisptr ))

#This is the same as __getitem__
#~     def at_index(self, unsigned int i):
#~         return wrap_Curve(self.thisptr.at_index(i))

    def front(self):
        """Get first curve."""
        #this is AFAIK the shortest way to do this
        cdef Curve * r = <Curve *> &self.thisptr.front()
        return wrap_Curve_p(r)

    def back(self):
        """Same as back_open."""
        cdef Curve * r = <Curve *> &self.thisptr.back()
        return wrap_Curve_p(r)

    def back_open(self):
        """Get last curve, treating self as open."""
        cdef Curve * r = <Curve *> &self.thisptr.back_open()
        return wrap_Curve_p(r)

    def back_closed(self):
        """Get last curve, treating self as closed."""
        cdef Curve * r = <Curve *> &self.thisptr.back_closed()
        return wrap_Curve_p(r)

    def back_default(self):
        """Get last curve."""
        cdef Curve * r = <Curve *> &self.thisptr.back_default()
        return wrap_Curve_p(r)

    def curves(self):
        """Same as curves_open"""
        return wrap_Iterator(self.thisptr.begin(), self.thisptr.begin_const(),  self.thisptr.end())

    def curves_open(self):
        """Return all curves as iterable, treating self as open."""
        return wrap_Iterator(self.thisptr.begin(), self.thisptr.begin_const(), self.thisptr.end_open())

    def curves_closed(self):
        """Return all curves as iterable, treating self as closed."""
        return wrap_Iterator(self.thisptr.begin(), self.thisptr.begin_const(), self.thisptr.end_closed())

    def curves_default(self):
        """Return all curves as iterable."""
        return wrap_Iterator(self.thisptr.begin(), self.thisptr.begin_const(), self.thisptr.end_default())

    def __iter__(self):
        return self.curves_default()

    def size_open(self):
        """Get number of curves, treating self as open."""
        return self.thisptr.size_open()

    def size_closed(self):
        """Get number of curves, treating self as closed."""
        return self.thisptr.size_closed()

    def size_default(self):
        """Get number of curves."""
        return self.thisptr.size_default()

    def size(self):
        """Same as size_open."""
        return self.thisptr.size()

#Does the same as size_open, which doesn't correspond with name.
#~     def max_size(self):
#~         return self.thisptr.max_size()

    def empty(self):
        """Test whether path contains no curves."""
        return self.thisptr.empty()

    def closed(self):
        """Return state of closed flag."""
        return self.thisptr.closed()

    def close(self, bint closed):
        """Set closed flag."""
        self.thisptr.close(closed)

    def bounds_fast(self):
        """Return fast bounding rectangle for path.

        It's not guaranteed to give the tighest bound.
        """
        return wrap_OptRect(self.thisptr.boundsFast())

    def bounds_exact(self):
        """Give the tighest bounding rectangle for path."""
        return wrap_OptRect(self.thisptr.boundsExact())

#~     def toPwSb(self):
#~         return wrap_Piecewise<Geom::D2<Geom::SBasis> >(self.thisptr.toPwSb())

    def point_at(self, double t):
        """Same as self(t)."""
        return wrap_Point(self.thisptr.pointAt(t))

    def value_at(self, double t, Dim2 d):
        """Same as self(t)[d]."""
        return self.thisptr.valueAt(t, d)

    def __call__(self, double t):
        """Evaluate path at time t.

        Equivalent to self[floor(t)](t-floor(t))
        """
        return wrap_Point(deref( self.thisptr ) (t) )

    def roots(self, double v, Dim2 d):
        """Find time values where self(t)[d] == v"""
        return wrap_vector_double(self.thisptr.roots(v, d))

    def all_nearest_times(self, cy_Point _point, double fr=-1, double to=1):
        """Return all values of t that |self(t) - point| is minimized."""
        if fr == -1:
            return wrap_vector_double(self.thisptr.allNearestTimes(deref( _point.thisptr )))
        return wrap_vector_double(self.thisptr.allNearestTimes(deref( _point.thisptr ), fr, to))


    def nearest_time_per_curve(self, cy_Point _point):
        """Find nearest points, return one time value per each curve."""
        return wrap_vector_double(self.thisptr.nearestTimePerCurve(deref( _point.thisptr )))

    def nearest_time(self, cy_Point _point, double fr=-1, double to=1):#, cy_double * distance_squared):
        """Return such t that |self(t) - point| is minimized."""
        if fr == -1:
            return self.thisptr.nearestTime(deref( _point.thisptr ), NULL)
        return self.thisptr.nearestTime(deref( _point.thisptr ), fr, to, NULL)

    def nearest_time_and_dist_sq(self, cy_Point _point, double fr=-1, double to=1):
        """Return such t that |self(t) - point| is minimized and square of that distance."""
        cdef double t, dist
        if fr == -1:
            t = self.thisptr.nearestTime(deref( _point.thisptr ), &dist )
        else:
            t = self.thisptr.nearestTime(deref( _point.thisptr ), fr, to, &dist)
        return (t, dist)

    def append_portion_to(self, cy_Path p, double f, double t):
        """Append portion of path to self."""
        self.thisptr.appendPortionTo(deref( p.thisptr ), f, t)

    def portion(self, Coord fr=0, Coord to=1, cy_Interval interval=None):
        """Return portion of curve between two time values.

        Alternatively use argument interval.
        """
        if interval is None:
            return wrap_Path(self.thisptr.portion(fr, to))
        else:
            return wrap_Path(self.thisptr.portion(deref( interval.thisptr )))

    def reversed(self):
        """Return reversed curve."""
        return wrap_Path(self.thisptr.reversed())

    def insert(self, int pos, curve, Stitching stitching=NO_STITCHING):
        """Insert curve into position pos.

        Args:
            pos: Position of inserted curve.
            curve: Curve to insert.
            stitching=NO_STITCHING
        """
        cdef Curve * cptr = get_Curve_p(curve)
        if cptr:
            self.thisptr.insert( self._iterator_at_index(pos), deref( cptr ), stitching )
        else:
            raise TypeError("passed curve is not C++ Curve")

    def insert_slice(self, int pos, cy_Path p, int first, int last, Stitching stitching=NO_STITCHING):
        """Insert curves to position pos.

        Args:
            pos: Position of inserted slice.
            p: Path from which slice is inserted.
            first: First inserted curve position (in p).
            last: Fist not inserted curve position (in p).
            stiching=NO_STITCHING
        """
        self.thisptr.insert(self._iterator_at_index(pos), p._const_iterator_at_index(first), p._const_iterator_at_index(last), stitching)

    def clear(self):
        """Clear all curves."""
        self.thisptr.clear()

    def erase(self, int pos, Stitching stitching=NO_STITCHING):
        """Erase curve at position pos.

        Args:
            pos: Position of erased curve.
            stitching=NO_STITCHING
        """
        self.thisptr.erase(self._iterator_at_index(pos), stitching)

    def erase_slice(self, int start, int end, Stitching stitching=NO_STITCHING):
        """Erase curves with indices [start, end).

        Args:
            start, end: Curves with indices start...end-1 are erased
            stitching=NO_STITCHING
        """
        self.thisptr.erase(self._iterator_at_index(start), self._iterator_at_index(end), stitching)

    def erase_last(self):
        """Erase last curve."""
        self.thisptr.erase_last()

    def replace(self, int replaced, curve, Stitching stitching=NO_STITCHING):
        """Replace curve at position replaced with another curve.

        Args:
            replaced: Position of replaced curve.
            curve: New curve.
            stitching=NO_STITCHING
        """
        cdef Curve * cptr = get_Curve_p(curve)
        if cptr:
            self.thisptr.replace(self._iterator_at_index(replaced), deref( cptr ), stitching)
        else:
            raise TypeError("passed curve is not C++ Curve")

    def replace_slice(self, int first_replaced, int last_replaced, curve, Stitching stitching=NO_STITCHING):
        """Replace slice of curves by new curve.

        Args:
            first_replaced, last_replace: Curves with indices
                first_replaced ... last_replaced
            curve: New curve.
            stitching=NO_STITCHING
        """
        cdef Curve * cptr = get_Curve_p(curve)
        if cptr:
            self.thisptr.replace(self._iterator_at_index(first_replaced), self._iterator_at_index(last_replaced), deref( cptr ), stitching)
        else:
            raise TypeError("passed curve is not C++ Curve")

#How to implement this nicely?
#~     def replaceByList(self, int replaced, cy_ConstIterator first, cy_ConstIterator last, Stitching stitching):
#~         self.thisptr.replace(deref( replaced.thisptr ), deref( first.thisptr ), deref( last.thisptr ), stitching)
#~     def replace(self, cy_Iterator first_replaced, cy_Iterator last_replaced, cy_ConstIterator first, cy_ConstIterator last, Stitching stitching):
#~         self.thisptr.replace(deref( first_replaced.thisptr ), deref( last_replaced.thisptr ), deref( first.thisptr ), deref( last.thisptr ), stitching)

    def start(self, cy_Point p):
        """Erase all curves and set first point."""
        self.thisptr.start(deref( p.thisptr ))

    def initial_point(self):
        """Get initial point."""
        return wrap_Point(self.thisptr.initialPoint())

    def final_point(self):
        """Get final point."""
        return wrap_Point(self.thisptr.finalPoint())

    def set_initial(self, cy_Point p):
        """Set initial point."""
        self.thisptr.setInitial(deref( p.thisptr ))

    def set_final(self, cy_Point p):
        """Set final point."""
        self.thisptr.setFinal(deref( p.thisptr ))

    def append_curve(self, curve, Stitching stitching=NO_STITCHING):
        """Append curve to path.

        Args:
            curve: Curve to append.
            stitching=NO_STITCHING
        """
        cdef Curve * cptr = get_Curve_p(curve)
        if cptr:
            self.thisptr.append( deref( cptr ), stitching)
        else:
            raise TypeError("passed curve is not C++ Curve")

    def append_SBasis(self, cy_SBasis x, cy_SBasis y, Stitching stitching=NO_STITCHING):
        """Append two SBasis functions to path.

        Args:
            x, y: SBasis functions to append.
            stitching=NO_STITCHING
        """
        cdef D2[SBasis] sb = D2[SBasis]( deref(x.thisptr), deref(y.thisptr) )
        self.thisptr.append(sb, stitching)

    def append_path(self, cy_Path other, Stitching stitching=NO_STITCHING):
        """Append another path to path.

        Args:
            other: Path to append.
            stitching=NO_STITCHING
        """
        self.thisptr.append(deref( other.thisptr ), stitching)

    def stitch_to(self, cy_Point p):
        """Set last point to p, creating stitching segment to it."""
        self.thisptr.stitchTo(deref( p.thisptr ))

cdef cy_Path wrap_Path(Path p):
    cdef Path * retp = new Path(Point())
    retp[0] = p
    cdef cy_Path r = cy_Path.__new__(cy_Path)
    r.thisptr = retp
    return r



from cython.operator cimport dereference as deref
from numbers import Number
from _cy_rectangle cimport cy_OptInterval, wrap_OptInterval, wrap_Rect, OptRect, wrap_OptRect
from _cy_rectangle cimport cy_Interval, wrap_Interval   

from _cy_affine cimport cy_Affine, wrap_Affine, get_Affine, is_transform

from _cy_curves cimport is_Curve, get_Curve_p
from _cy_path cimport wrap_Path


cdef class cy_Circle:
    
    """Circle in 2D plane.
    
    Corresponds to Circle class in 2geom.
    """

    cdef Circle* thisptr

    def __cinit__(self, cy_Point center=None, double r=0):
        """Create circle from center and radius."""
        if center is None:
            self.thisptr = new Circle()
        else:
            self.thisptr = new Circle(deref( center.thisptr ), r)

    @classmethod
    def from_coefficients(self, double A, double B, double C, double D):
        """Create circle form implicit equation coefficients:
        
        Implicit equation is Ax**2 + Ay**2 + Bx + Cy + D = 0
        """
        return wrap_Circle( Circle(A, B, C, D) )

    @classmethod
    def from_points(self, points):
        """Create best fitting circle from at least three points."""
        return wrap_Circle( Circle( make_vector_point(points) ) )

    def set_center(self, cy_Point c):
        """Set coordinates of center."""
        self.thisptr.setCenter(deref(c.thisptr))

    def set_radius(self, double r):
        """Set the circle's radius."""
        self.thisptr.setRadius(r)

    def set_coefficients(self, double A, double B, double C, double D):
        """Set implicit equation coefficients:
        
        Implicit equation is Ax**2 + Ay**2 + Bx + Cy + D = 0
        """
        self.thisptr.set(A, B, C, D)
    
    def fit(self, points):
        """Set circle to the best fit of at least three points."""
        self.thisptr.fit( make_vector_point(points) )
        
    def arc(self, cy_Point initial, cy_Point inner, cy_Point final, bint _svg_compliant=True):
        """Get (SVG)EllipticalArc.
        
        Args:
            initial: Initial point of arc
            inner: Inner point of arc.
            final: Final point of arc.
        """
        return wrap_EllipticalArc( deref(self.thisptr.arc(deref( initial.thisptr ), deref( inner.thisptr ), deref( final.thisptr ))) )

    def center(self):
        """Get center of circle in point."""
        return wrap_Point(self.thisptr.center())

    def radius(self):
        """Get radius of circle."""
        return self.thisptr.radius()

cdef cy_Circle wrap_Circle(Circle p):
    cdef Circle * retp = new Circle()
    retp[0] = p
    cdef cy_Circle r = cy_Circle.__new__(cy_Circle)
    r.thisptr = retp
    return r

cdef class cy_Ellipse:
    """Ellipse in 2D plane.
    
    Corresponds to Ellipse class in 2geom.
    """
    
    cdef Ellipse* thisptr
    
    def __cinit__(self, cy_Point center=None, rx=0, ry=0, double a=0):
        """Create new ellipse:
        
        Args:
            center: Center of ellipse (between foci)
            rx, ry: major and minor semi-axis.
            a: angle of major axis.
        """
        if center is None:
            self.thisptr = new Ellipse()
        else:
            self.thisptr = new Ellipse(center.x, center.y, rx, ry, a)
    
    @classmethod
    def from_coefficients(cls, double A, double B, double C, double D, double E, double F):
        """Create ellipse from coefficients of implicit equation.
        
        Implicit equation has form Ax**2 + Bxy + Cy**2 + Dx + Ey + F = 0
        """
        return wrap_Ellipse(Ellipse(A, B, C, D, E, F))
    
    @classmethod
    def from_circle(cls, cy_Circle c):
        """Create ellipse identical to circle."""
        return wrap_Ellipse(Ellipse(deref( c.thisptr )))
    
    @classmethod
    def from_points(cls, points):
        """Create ellipse fitting at least 5 points."""
        return wrap_Ellipse( Ellipse( make_vector_point(points) ) )
    
    def set(self, cy_Point center, double rx, double ry, double a):
        """Set center, rays and angle of ellipse.
        
        Args:
            center: Center of ellipse.
            rx, ry: Major and minor semi-axis.
            a: angle of major axis.
        self.thisptr.set(center.x, center.y, rx, ry, a)
        """

    def set_coefficients(self, double A, double B, double C, double D, double E, double F):
        """Set coefficients of implicit equation.
        
        Implicit equation has form Ax**2 + Bxy + Cy**2 + Dx + Ey + F = 0
        """
        self.thisptr.set(A, B, C, D, E, F)

    def set_points(self, points):
        """Set ellipse to the best fit of at least five points."""
        self.thisptr.set( make_vector_point(points) )

    def arc(self, cy_Point initial, cy_Point inner, cy_Point final, bint svg_compliant=True):
        """Get (SVG)EllipticalArc.
        
        Args:
            initial: Initial point of arc
            inner: Inner point of arc.
            final: Final point of arc.
        """
        return wrap_EllipticalArc( deref(self.thisptr.arc(deref( initial.thisptr ), deref( inner.thisptr ), deref( final.thisptr ))) )

    def center(self):
        """Get center of ellipse."""
        return wrap_Point(self.thisptr.center())

    def ray(self, Dim2 d):
        """Get major/minor semi-axis."""
        return self.thisptr.ray(d)

    def rot_angle(self):
        """Get angle of major axis."""
        return self.thisptr.rot_angle()

    def implicit_form_coefficients(self):
        """Get coefficients of implicit equation in list."""
        return wrap_vector_double(self.thisptr.implicit_form_coefficients())

    def transformed(self, m):
        """Return transformed ellipse."""
        cdef Affine at
        if is_transform(m):
            at = get_Affine(m)
        return wrap_Ellipse(self.thisptr.transformed(at))
        
cdef cy_Ellipse wrap_Ellipse(Ellipse p):
    cdef Ellipse * retp = new Ellipse()
    retp[0] = p
    cdef cy_Ellipse r = cy_Ellipse.__new__(cy_Ellipse)
    r.thisptr = retp
    return r



from _common_decl cimport *

from _cy_rectangle cimport Rect, cy_Rect, wrap_Rect
from _cy_primitives cimport Point, cy_Point, wrap_Point


cdef extern from "2geom/affine.h" namespace "Geom":
    cdef cppclass Affine:

        Affine(Affine &)
        Affine()
        Affine(Coord, Coord, Coord, Coord, Coord, Coord)

        Coord operator[](unsigned int)

        Affine & operator*(Affine &)
        Affine & operator*(Translate &)
        Affine & operator*(Scale &)
        Affine & operator*(Rotate &)
        Affine & operator*(HShear &)
        Affine & operator*(VShear &)
        Affine & operator*(Zoom &)

        bint operator==(Affine &)
        bint operator!=(Affine &)

        Point xAxis()
        Point yAxis()
        Point translation()
        Coord expansionX()
        Coord expansionY()
        Point expansion()

        void setXAxis(Point &)
        void setYAxis(Point &)
        void setTranslation(Point &)
        void setExpansionX(Coord)
        void setExpansionY(Coord)
        void setIdentity()

        bint isIdentity(Coord)
        bint isTranslation(Coord)
        bint isScale(Coord)
        bint isUniformScale(Coord)
        bint isRotation(Coord)
        bint isHShear(Coord)
        bint isVShear(Coord)
        bint isNonzeroTranslation(Coord)
        bint isNonzeroScale(Coord)
        bint isNonzeroUniformScale(Coord)
        bint isNonzeroRotation(Coord)
        bint isNonzeroHShear(Coord)
        bint isNonzeroVShear(Coord)
        bint isZoom(Coord)
        bint preservesArea(Coord)
        bint preservesAngles(Coord)
        bint preservesDistances(Coord)
        bint flips()
        bint isSingular(Coord)

        Affine withoutTranslation()
        Affine inverse()

        Coord det()
        Coord descrim2()
        Coord descrim()

    bint are_near(Affine &, Affine &, Coord)
    Affine a_identity "Geom::Affine::identity" ()

cdef extern from "2geom/transforms.h" namespace "Geom":
    Affine reflection(Point &, Point &)
    #TODO find out how cython __pow__ works
    Affine      pow(Affine &, int)
    Translate   pow(Translate &, int)
    Scale       pow(Scale &, int)
    Rotate      pow(Rotate &, int)
    HShear      pow(HShear &, int)
    VShear      pow(VShear &, int)
    Zoom        pow(Zoom &, int)

cdef class cy_Affine:
    cdef Affine* thisptr

cdef cy_Affine wrap_Affine(Affine)

#helper functions
cdef Affine get_Affine(t)
cdef bint is_transform(t)


cdef extern from "2geom/transforms.h" namespace "Geom":
    cdef cppclass Translate:
        Translate(Translate &)
        Translate()
        Translate(Point &)
        Translate(Coord, Coord)
        Coord operator[](Dim2)
        Coord operator[](unsigned int)
        Translate & operator*(Translate &)
        Affine & operator*(Affine &)
        bint operator==(Translate &)
        bint operator!=(Translate &)

        Affine operator()

        Point vector()
        Translate inverse()

    Translate t_identity "Geom::Translate::identity" ()

cdef class cy_Translate:
    cdef Translate* thisptr


cdef extern from "2geom/transforms.h" namespace "Geom":
    cdef cppclass Scale:
        Scale(Scale &)
        Scale()
        Scale(Point &)
        Scale(Coord, Coord)
        Scale(Coord)
        Coord operator[](Dim2)
        Scale & operator*(Scale &)
        Affine & operator*(Affine &)
        bint operator==(Scale &)
        bint operator!=(Scale &)

        Affine operator()

        Point vector()
        Scale inverse()
        Scale identity()

    Scale s_identity "Geom::Scale::identity" ()

cdef class cy_Scale:
    cdef Scale* thisptr


cdef extern from "2geom/transforms.h" namespace "Geom":
    cdef cppclass Rotate:
        Rotate(Rotate &)
        Rotate()
        Rotate(Coord)
        Rotate(Point &)
        Rotate(Coord, Coord)
        Point vector()

        Coord operator[](Dim2)
        Coord operator[](unsigned int)
        Rotate & operator*(Rotate &)
        Affine & operator*(Affine &)
        bint operator==(Rotate &)
        bint operator!=(Rotate &)

        Affine operator()
        Rotate inverse()

    Rotate r_identity "Geom::Rotate::identity" ()

cdef extern from "2geom/transforms.h" namespace "Geom::Rotate":
    Rotate from_degrees(Coord)


cdef class cy_Rotate:
    cdef Rotate* thisptr

cdef extern from "2geom/transforms.h" namespace "Geom":
    cdef cppclass VShear:
        VShear(VShear &)
        VShear(Coord)
        Coord factor()
        void setFactor(Coord)

        VShear &operator*(VShear)
        Affine & operator*(Affine &)
        bint operator==(VShear &)
        bint operator!=(VShear &)
        Affine operator()

        VShear inverse()

    VShear vs_identity "Geom::VShear::identity"()

cdef class cy_VShear:
    cdef VShear* thisptr


cdef extern from "2geom/transforms.h" namespace "Geom":
    cdef cppclass HShear:
        HShear(HShear &)
        HShear(Coord)
        Coord factor()
        void setFactor(Coord)
        HShear &operator*(HShear)
        Affine & operator*(Affine &)
        bint operator==(HShear &)
        bint operator!=(HShear &)
        Affine operator()

        HShear inverse()

    HShear hs_identity "Geom::HShear::identity"()

cdef class cy_HShear:
    cdef HShear* thisptr


cdef extern from "2geom/transforms.h" namespace "Geom":
    cdef cppclass Zoom:
        Zoom(Zoom &)
        Zoom(Coord)
        Zoom(Translate &)
        Zoom(Coord, Translate &)

        Zoom & operator*(Zoom &)
        Affine & operator*(Affine &)
        bint operator==(Zoom &)
        bint operator!=(Zoom &)

        Affine operator()

        Coord scale()
        void setScale(Coord)
        Point translation()
        void setTranslation(Point &)

        Zoom inverse()

        Zoom()

    Zoom z_identity "Geom::Zoom::identity" ()

cdef extern from "2geom/transforms.h" namespace "Geom::Zoom":
    Zoom map_rect(Rect &, Rect &)

cdef class cy_Zoom:
    cdef Zoom* thisptr


cdef extern from "2geom/affine.h" namespace "Geom":
    cdef cppclass Eigen:
        Point *vectors
        double *values
        Eigen(Affine &)
        Eigen(double[2][2])



from _common_decl cimport *

from libcpp.vector cimport vector
from libcpp.pair cimport pair

from _cy_rectangle cimport Interval, OptInterval, Rect, OptRect
from _cy_affine cimport Affine
from _cy_primitives cimport Point, cy_Point, wrap_Point, wrap_vector_point, make_vector_point
from _cy_primitives cimport Angle, cy_Angle, wrap_Angle
from _cy_primitives cimport AngleInterval


cdef extern from "2geom/d2.h" namespace "Geom":
    cdef cppclass D2[T]:
        D2()
        D2(T &, T &)
        T& operator[](unsigned i)

cdef extern from "2geom/curve.h" namespace "Geom":
    cdef cppclass Curve:
        Curve()
        Point initialPoint()
        Point finalPoint()
        bint isDegenerate()
        Point pointAt(Coord)
        Coord valueAt(Coord, Dim2)
        Point operator()(Coord)
        vector[Point] pointAndDerivatives(Coord, unsigned int)
        void setInitial(Point &)
        void setFinal(Point &)
        Rect boundsFast()
        Rect boundsExact()
        OptRect boundsLocal(OptInterval &, unsigned int)
        OptRect boundsLocal(OptInterval &)
        Curve * duplicate()
        Curve * transformed(Affine &)
        Curve * portion(Coord, Coord)
        Curve * portion(Interval &)
        Curve * reverse()
        Curve * derivative()
        Coord nearestTime(Point &, Coord, Coord)
        Coord nearestTime(Point &, Interval &)
        vector[double] allNearestTimes(Point &, Coord, Coord)
        vector[double] allNearestTimes(Point &, Interval &)
        Coord length(Coord)
        vector[double] roots(Coord, Dim2)
        int winding(Point &)
        Point unitTangentAt(Coord, unsigned int)
        D2[SBasis] toSBasis()
        int degreesOfFreedom()
        bint operator==(Curve &)

cdef class cy_Curve:
    cdef Curve* thisptr

#~ cdef cy_Curve wrap_Curve(Curve & p)
cdef cy_Curve wrap_Curve_p(Curve * p)


cdef extern from "2geom/linear.h" namespace "Geom":
    cdef cppclass Linear:
#~         Linear(Linear &)
        Linear()
        Linear(double, double)
        Linear(double)
        double operator[](int const)
        bint isZero(double)
        bint isConstant(double)
        bint isFinite()
        double at0()
        double at1()
        double valueAt(double)
        double operator()(double)
        SBasis toSBasis()
        OptInterval bounds_exact()
        OptInterval bounds_fast()
        OptInterval bounds_local(double, double)
        double tri()
        double hat()

    bint operator==(Linear &, Linear &)
    bint operator!=(Linear &, Linear &)
    Linear operator*(Linear &, double)
    Linear operator+(Linear &, double)
    Linear operator+(Linear &, Linear &)
    #cython has trouble resolving these
    Linear L_neg "operator-" (Linear &)
    Linear L_sub_Ld "operator-"(Linear &, double)
    Linear L_sub_LL "operator-"(Linear &, Linear &)
    Linear operator/(Linear &, double)

    double lerp(double, double, double)
    Linear reverse(Linear &)


cdef extern from "2geom/sbasis.h" namespace "Geom":
    cdef cppclass SBasis:

        SBasis()
        SBasis(double)
        SBasis(double, double)
        SBasis(SBasis &)
        SBasis(vector[Linear] &)
        SBasis(Linear &)

        size_t size()
        Linear operator[](unsigned int)
        bint empty()
        Linear & back()
        void pop_back()
        void resize(unsigned int)
        void resize(unsigned int, Linear &)
        void reserve(unsigned int)
        void clear()
#~         void insert(::__gnu_cxx::__normal_iterator<Geom::Linear*, std::vector<Geom::Linear, std::allocator<Geom::Linear> > >, ::__gnu_cxx::__normal_iterator<Geom::Linear const*, std::vector<Geom::Linear, std::allocator<Geom::Linear> > >, ::__gnu_cxx::__normal_iterator<Geom::Linear const*, std::vector<Geom::Linear, std::allocator<Geom::Linear> > >)
        Linear & at(unsigned int)
        bint operator==(SBasis &)
        bint operator!=(SBasis &)

        bint isZero(double)
        bint isConstant(double)
        bint isFinite()
        double at0()
        double at1()
        int degreesOfFreedom()
        double valueAt(double)
        double operator()(double)
        vector[double] valueAndDerivatives(double, unsigned int)
        SBasis toSBasis()
        double tailError(unsigned int)
        SBasis operator()(SBasis &)
        void normalize()
        void truncate(unsigned int)

    SBasis operator*(SBasis &, SBasis &)
    SBasis operator*(double, SBasis &)
    SBasis operator*(SBasis &, double)
    SBasis operator+(SBasis &, double)
    SBasis operator+(SBasis &, SBasis &)
    SBasis SB_sub_Sd "operator-" (SBasis &, double)
    SBasis SB_sub_SS "operator-" (SBasis &, SBasis &)
    SBasis SB_neg "operator-" (SBasis &)
    SBasis operator/(SBasis &, double)

    SBasis sin(Linear, int)
    SBasis cos(Linear, int)
    SBasis sqrt(SBasis &, int)
    SBasis reciprocal(Linear &, int)
    SBasis shift(Linear &, int)
    SBasis shift(SBasis &, int)
    SBasis inverse(SBasis, int)

    SBasis portion(SBasis &, Interval)
    SBasis portion(SBasis &, double, double)
    SBasis compose(SBasis &, SBasis &, unsigned int)
    SBasis compose(SBasis &, SBasis &)
    SBasis truncate(SBasis &, unsigned int)

    unsigned int valuation(SBasis &, double)

    vector[ vector[double] ] multi_roots(SBasis &, vector[double] &, double, double, double, double) #TODO
    vector[Interval] level_set(SBasis &, Interval &, double, double, double)
    vector[Interval] level_set(SBasis &, double, double, double, double, double)
    vector[double] roots(SBasis &, Interval const)
    vector[double] roots(SBasis &)
    vector[ vector[Interval] ] level_sets(SBasis &, vector[Interval] &, double, double, double) #TODO
    vector[ vector[Interval] ] level_sets(SBasis &, vector[double] &, double, double, double, double) #TODO

    SBasis reverse(SBasis &)
    SBasis derivative(SBasis &)
    SBasis integral(SBasis &)
    SBasis divide(SBasis &, SBasis &, int)
    SBasis compose_inverse(SBasis &, SBasis &, unsigned int, double)
    SBasis multiply(SBasis &, SBasis &)
    SBasis multiply_add(SBasis &, SBasis &, SBasis)

    OptInterval bounds_exact(SBasis &)
    OptInterval bounds_local(SBasis &, OptInterval &, int)
    OptInterval bounds_fast(SBasis &, int)

cdef class cy_SBasis:
    cdef SBasis* thisptr

cdef extern from "2geom/sbasis-curve.h" namespace "Geom":
    cdef cppclass SBasisCurve:
        SBasisCurve(D2[SBasis] &)
        SBasisCurve(Curve &)
        Curve * duplicate()
        Point initialPoint()
        Point finalPoint()
        bint isDegenerate()
        Point pointAt(Coord)
        Point operator()(double)
        vector[Point] pointAndDerivatives(Coord, unsigned int)
        Coord valueAt(Coord, Dim2)
        void setInitial(Point &)
        void setFinal(Point &)
        Rect boundsFast()
        Rect boundsExact()
        OptRect boundsLocal(OptInterval &, unsigned int)
        vector[double] roots(Coord, Dim2)
        Coord nearestTime(Point &, Coord, Coord)
        vector[double] allNearestTimes(Point &, Coord, Coord)
        Coord length(Coord)
        Curve * portion(Coord, Coord)
        Curve * transformed(Affine &)
        Curve * derivative()
        D2[SBasis] toSBasis()
        int degreesOfFreedom()


cdef extern from "2geom/bezier.h" namespace "Geom":
    cdef cppclass Bezier:
#~         struct Order
#~             pass
        unsigned int order()
        unsigned int size()
        Bezier()
#~         Bezier(Bezier.Order)
        Bezier(Coord)
        Bezier(Coord, Coord)
        Bezier(Coord, Coord, Coord)
        Bezier(Coord, Coord, Coord, Coord)
        void resize(unsigned int, Coord)
        void clear()
        unsigned int degree()
        bint isZero(double)
        bint isConstant(double)
        bint isFinite()
        Coord at0()
        Coord at1()
        Coord valueAt(double)
        Coord operator()(double)
        SBasis toSBasis()
        Coord & operator[](unsigned int)
        void setPoint(unsigned int, double)
        vector[double] valueAndDerivatives(Coord, unsigned int)
        pair[Bezier, Bezier] subdivide(Coord)
        vector[double] roots()
        vector[double] roots(Interval const)
        Bezier forward_difference(unsigned int)
        Bezier elevate_degree()
        Bezier reduce_degree()
        Bezier elevate_to_degree(unsigned int)
        Bezier deflate()
        Bezier(Coord *, unsigned int)

    Bezier operator*(Bezier &, double)
    Bezier operator+(Bezier &, double)
    Bezier operator-(Bezier &, double)
    Bezier operator/(Bezier &, double)

    Bezier portion(Bezier &, double, double)
    OptInterval bounds_local(Bezier &, OptInterval)
    Coord casteljau_subidivision(Coord, Coord *, Coord *, Coord *, unsigned int)
    void bezier_to_sbasis(SBasis &, Bezier &)
    Bezier integral(Bezier &)
    Bezier derivative(Bezier &)
    OptInterval bounds_exact(Bezier &)
    double bernsteinValueAt(double, double *, unsigned int)
    vector[Point] bezier_points(D2[Bezier] &)
    OptInterval bounds_fast(Bezier &)
    Bezier multiply(Bezier &, Bezier &)
    Bezier reverse(Bezier &)

#This is ugly workaround around cython's lack of support for integer template parameters
cdef extern from *:
    ctypedef int n_0 "0"
    ctypedef int n_1 "1"

cdef extern from "2geom/bezier-curve.h" namespace "Geom":
    cdef cppclass BezierCurve:
        unsigned int order()
        vector[Point] controlPoints()
        void setPoint(unsigned int, Point)
        void setPoints(vector[Point] &)
        Point operator[](unsigned int)
        Point initialPoint()
        Point finalPoint()
        bint isDegenerate()
        void setInitial(Point &)
        void setFinal(Point &)
        Rect boundsFast()
        Rect boundsExact()
        OptRect boundsLocal(OptInterval &, unsigned int)
        Curve * duplicate()
        Curve * portion(Coord, Coord)
        Curve * reverse()
        Curve * transformed(Affine &)
        Curve * derivative()
        int degreesOfFreedom()
        vector[double] roots(Coord, Dim2)
        Coord length(Coord)
        Point pointAt(Coord)
        vector[Point] pointAndDerivatives(Coord, unsigned int)
        Coord valueAt(Coord, Dim2)
        D2[SBasis] toSBasis()
        #protected:
#~         BezierCurve()
#~         BezierCurve(D2[Bezier] &)
#~         BezierCurve(Bezier &, Bezier &)
#~         BezierCurve(vector[Point] &)
    #~ Point middle_point(LineSegment &)
    #~ Coord length(LineSegment &)
    Coord bezier_length(Point, Point, Point, Point, Coord)
    Coord bezier_length(Point, Point, Point, Coord)
    Coord bezier_length(vector[Point] &, Coord)

    cdef cppclass LineSegment:
        LineSegment()
        LineSegment(D2[Bezier] &)
        LineSegment(Bezier, Bezier)
        LineSegment(vector[Point] &)
        LineSegment(Point, Point)
        pair[LineSegment, LineSegment] subdivide(Coord)

        Curve * duplicate()
        Curve * reverse()
        Curve * transformed(Affine &)
        Curve * derivative()

    cdef cppclass QuadraticBezier:
        QuadraticBezier()
        QuadraticBezier(D2[Bezier] &)
        QuadraticBezier(Bezier, Bezier)
        QuadraticBezier(vector[Point] &)
        QuadraticBezier(Point, Point, Point)
        pair[QuadraticBezier, QuadraticBezier] subdivide(Coord)

        Curve * duplicate()
        Curve * reverse()
        Curve * transformed(Affine &)
        Curve * derivative()


    cdef cppclass CubicBezier:
        CubicBezier()
        CubicBezier(D2[Bezier] &)
        CubicBezier(Bezier, Bezier)
        CubicBezier(vector[Point] &)
        CubicBezier(Point, Point, Point, Point)
        pair[CubicBezier, CubicBezier] subdivide(Coord)

        Curve * duplicate()
        Curve * reverse()
        Curve * transformed(Affine &)
        Curve * derivative()

#~     cdef cppclass BezierCurveN[n_1]:
#~         BezierCurveN ()
#~         BezierCurveN(Bezier &, Bezier &)
#~         #BezierCurveN(Point &, Point &)
#~         BezierCurveN(vector[Point] &)
#~         pair[BezierCurveN, BezierCurveN] subdivide(Coord)
#~
#~         Curve * duplicate()
#~         Curve * reverse()
#~         Curve * transformed(Affine &)
#~         Curve * derivative()

cdef class cy_BezierCurve:
    cdef BezierCurve* thisptr

cdef class cy_LineSegment(cy_BezierCurve):
    pass

cdef cy_LineSegment wrap_LineSegment(LineSegment p)

cdef extern from "2geom/bezier-curve.h" namespace "Geom::BezierCurve":
    BezierCurve * create(vector[Point] &)

cdef extern from "2geom/elliptical-arc.h" namespace "Geom":
    cdef cppclass EllipticalArc:
        EllipticalArc(EllipticalArc &)
        EllipticalArc()
        EllipticalArc(Point, Coord, Coord, Coord, bint, bint, Point)
        Interval angleInterval()
        Angle rotationAngle()
        Coord ray(Dim2)
        Point rays()
        bint largeArc()
        bint sweep()
        LineSegment chord()
        void set(Point &, double, double, double, bint, bint, Point &)
        void setExtremes(Point &, Point &)
        Coord center(Dim2)
        Point center()
        Coord sweepAngle()
        bint containsAngle(Coord)
        Point pointAtAngle(Coord)
        Coord valueAtAngle(Coord, Dim2)
        Affine unitCircleTransform()
        bint isSVGCompliant()
        pair[EllipticalArc, EllipticalArc] subdivide(Coord)
        Point initialPoint()
        Point finalPoint()
        Curve * duplicate()
        void setInitial(Point &)
        void setFinal(Point &)
        bint isDegenerate()
        Rect boundsFast()
        Rect boundsExact()
        OptRect boundsLocal(OptInterval &, unsigned int)
        vector[double] roots(double, Dim2)
        double nearestTime(Point &, double, double)
        int degreesOfFreedom()
        Curve * derivative()
        Curve * transformed(Affine &)
        vector[Point] pointAndDerivatives(Coord, unsigned int)
        D2[SBasis] toSBasis()
        double valueAt(Coord, Dim2)
        Point pointAt(Coord)
        Curve * portion(double, double)
        Curve * reverse()

cdef class cy_EllipticalArc:
    cdef EllipticalArc* thisptr
cdef cy_EllipticalArc wrap_EllipticalArc(EllipticalArc p)

cdef Curve * get_Curve_p(object c)
cdef bint is_Curve(object c)



import gdb.printing
import math

def xstr(val):
    if val.is_optimized_out:
        return "<optimized out>"
    else:
        return str(val)

class GeomPointPrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return "(" + xstr(self.val["_pt"][0]) + ", " + xstr(self.val["_pt"][1]) + ")"

class GeomIntervalPrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return "[" + xstr(self.val["_b"][0]) + ", " + xstr(self.val["_b"][1]) + ")"

class GeomOptIntervalPrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        if self.val["m_initialized"]:
            addr = self.val["m_storage"]["dummy_"]["data"].address
            if self.val.type == gdb.lookup_type('Geom::OptInterval'):
                return str(addr.cast(gdb.lookup_type('Geom::Interval').pointer()).dereference())
            else:
                return str(addr.cast(gdb.lookup_type('Geom::GenericInterval<int>').pointer()).dereference())
        else:
            return "empty"

class GeomRectPrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return str(self.val["f"][0]) + " x " + str(self.val["f"][1])

class GeomOptRectPrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        if self.val["m_initialized"]:
            addr = self.val["m_storage"]["dummy_"]["data"].address
            if self.val.type == gdb.lookup_type('Geom::OptRect'):
                return str(addr.cast(gdb.lookup_type('Geom::Rect').pointer()).dereference())
            else:
                return str(addr.cast(gdb.lookup_type('Geom::GenericRect<int>').pointer()).dereference())
        else:
            return "empty"

class GeomAffinePrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return ("\n | %8f  %8f  0 |" +
                "\n | %8f  %8f  0 |" +
                "\n | %8f  %8f  1 |") % tuple(self.val["_c"][x] for x in range(0,6))

class GeomTranslatePrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return "translate by " + str(self.val["vec"])

class GeomScalePrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return "scale by " + str(self.val["vec"])

class GeomRotatePrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        x = self.val["vec"][0]
        y = self.val["vec"][1]
        angle = math.atan2(y, x) / math.pi * 180.0
        return "rotate by %s = %f degrees" % (str(self.val["vec"]), angle)

class GeomZoomPrinter:
    def __init__(self, val):
        self.val = val
    def to_string(self):
        return "zoom %f%% at %s" % (self.val["_scale"] * 100, str(self.val["_trans"]))


def lib2geom_pretty_printer():
	pp = gdb.printing.RegexpCollectionPrettyPrinter("lib2geom")
	pp.add_printer('Geom::Point', '^Geom::(Int)?Point$', GeomPointPrinter)
	pp.add_printer('Geom::Interval', '^(Geom::Interval|Geom::GenericInterval<int>)$', GeomIntervalPrinter)
	pp.add_printer('Geom::OptInterval', '^(Geom::OptInterval|Geom::GenericOptInterval<int>)$', GeomOptIntervalPrinter)
	pp.add_printer('Geom::Rect', '^(Geom::Rect|Geom::GenericRect<int>)$', GeomRectPrinter)
	pp.add_printer('Geom::OptRect', '^(Geom::OptRect|Geom::GenericOptRect<int>)$', GeomOptRectPrinter)
	pp.add_printer('Geom::Affine', '^Geom::Affine$', GeomAffinePrinter)
	pp.add_printer('Geom::Translate', '^Geom::Translate$', GeomTranslatePrinter)
	pp.add_printer('Geom::Scale', '^Geom::Scale', GeomScalePrinter)
	pp.add_printer('Geom::Rotate', '^Geom::Rotate$', GeomRotatePrinter)
	pp.add_printer('Geom::Zoom', '^Geom::Zoom$', GeomZoomPrinter)
	return pp

def register_lib2geom_printers():
	gdb.printing.register_pretty_printer(
		gdb.current_objfile(),
		lib2geom_pretty_printer())



#!/usr/bin/python

import gtk,math
import pangocairo,cairo
import gobject

# def draw_text(cr, pos, txt, bottom = False):
# def draw_number(cr, pos, num):

def draw_circ(cr, (x, y)):
    cr.new_sub_path()
    cr.arc(x, y, 3, 0, math.pi*2)
    cr.stroke()
def draw_cross(cr, (x, y)):
    cr.move_to(x-3, y-3)
    cr.line_to(x+3, y+3)
    cr.move_to(x-3, y+3)
    cr.line_to(x+3, y-3)
    cr.stroke()

class Handle:
    def __init__(self):
        pass
    def draw(self, cr, annotes):
        pass
    def hit(self, (x, y)):
        return None
    def move_to(self, hit, om, m):
        pass
    def scroll(self, pos, dir):
        pass

class PointHandle(Handle):
    def __init__(self, x, y, name=""):
        Handle.__init__(self)
        self.pos = (x,y)
        self.name = name
    def draw(self, cr, annotes):
        draw_circ(cr, self.pos)
        if annotes:
            draw_text(cr, self.pos, str(self.name))
    def hit(self, mouse):
        if math.hypot(mouse[0] - self.pos[0], mouse[1] - self.pos[1]) < 5:
            return 0
        return None
    def move_to(self, hit, om, m):
        self.pos = m

class PointSetHandle(Handle):
    def __init__(self, pts=None, name=""):
        Handle.__init__(self)
        self.pts = pts or []
        self.name = name
    def draw(self, cr, annotes):
        for p in self.pts:
            draw_circ(cr, p)
            if annotes:
                draw_text(cr, p, str(self.name))
    def hit(self, mouse):
        for i,p in enumerate(self.pts):
            if math.hypot(mouse[0] - p[0], mouse[1] - p[1]) < 5:
                return i
        return None
    def move_to(self, hit, om, m):
        self.pts[hit[1]] = m
    def append(self, x,y):
        self.pts.append((x,y))

class Toy:
    def __init__(self):
        self.handles = []
        self.mouse_down = False
        self.old_mouse = None
        self.selected = None
        self.notify = "notify"
        self.origin = [0,0]
        self.interactive_level = 0
        self.transform = None

    def draw(self, cr, (width, height), save):
        bounds = self.should_draw_bounds()
        self.transform = cr.get_matrix()
        self.transform.invert()
        if bounds == 1:
            cr.set_source_rgba(0., 0., 0, 0.8)
            cr.set_line_width (0.5)
            for i in [1,3]:
                cr.move_to(0, i*width/4)
                cr.line_to(width, i*width/4)
                cr.move_to(i*width/4, 0)
                cr.line_to(i*width/4, height)
        elif bounds == 2:
            cr.set_source_rgba (0., 0., 0, 0.8)
            cr.set_line_width (0.5)
            cr.move_to(0, width/2)
            cr.line_to(width, width/2)
            cr.move_to(width/2, 0)
            cr.line_to(width/2, height)

        cr.set_source_rgba (0., 0.5, 0, 1)
        cr.set_line_width (1)
        annotes = self.should_draw_numbers()
        for i,h in enumerate(self.handles):
            cr.save()
            if self.selected and i == self.selected[0]:
                cr.set_source_rgba (0.5, 0, 0, 1)
            h.draw(cr, annotes)
            cr.restore()

        cr.set_source_rgba (0., 0.5, 0, 0.8)
        if self.notify:
            cr.save()
            cr.identity_matrix()
            bnds = draw_text(cr, (0, height), self.notify, True)
            l,t,w,h = bnds[1]
            cr.set_source_rgba(1,1,1,0.9)
            cr.rectangle(l,t+height-h, w, h)
            cr.fill()
            cr.set_source_rgba(0,0,0,1)
            draw_text(cr, (0, height), self.notify, True)
            cr.restore()
        
    def mouse_moved(self, e):
        mouse = (e.x, e.y)
        if self.transform != None:
            mouse = self.transform.transform_point(e.x, e.y)
    
        if e.state & (gtk.gdk.BUTTON1_MASK | gtk.gdk.BUTTON3_MASK):
            self.interactive_level = 2
            if self.selected:
                self.handles[self.selected[0]].move_to(self.selected, self.old_mouse, mouse)
        self.old_mouse = mouse
        self.canvas.queue_draw()
        #self.redraw()
    def mouse_pressed(self, e):
        self.interactive_level = 1
        mouse = (e.x, e.y)
        if self.transform != None:
            mouse = self.transform.transform_point(e.x, e.y)
        if e.button == 1:
            for i,h in enumerate(self.handles):
                hit = h.hit(mouse)
                if hit != None:
                    self.selected = (i, hit)
            self.mouse_down = True
        self.old_mouse = mouse
        self.canvas.queue_draw()
        #self.redraw()

    def mouse_released(self, e):
        self.interactive_level = 1
        self.selected = None
        if e.button == 1:
            self.mouse_down = False
        self.canvas.queue_draw()
        #self.redraw()

    def scroll_event(self, da, ev):
        self.interactive_level = 1
        #print 'scroll:'+'\n'.join([str((x, getattr(ev, x))) for x in dir(ev)])
        #print 'end'
        da.queue_draw()
        hit = None
        for i,h in enumerate(self.handles):
            hit = h.scroll((ev.x,ev.y), ev.direction)
            if hit != None:
                break
        return hit != None
        
    def key_hit(self, e):
        pass

    def should_draw_numbers(self):
        return True
    def should_draw_bounds(self):
        return 0
    
    def first_time(self, argv):
        pass
    
    def gtk_ready(self):
        pass
    
    def resize_canvas(self, s):
        pass
    def redraw(self):
        self.window.queue_draw()
    def get_save_size(self):
        return (0,0)+self.window.window.get_size()
    def delete_event(self, window, e):
        gtk.main_quit()
        return False
    
    def expose_event(self, widget, event):
        cr = widget.window.cairo_create()

        width, height = widget.window.get_size()
        global resized

        if not resized:
            alloc_size = ((0, width), (0, height))
            self.resize_canvas(alloc_size)
            resized = True
        cr.translate(self.origin[0], self.origin[1])
        self.draw(cr, (width, height), False)

        return True

    def mouse_motion_event(self, widget, e):
        e.x -= self.origin[0]
        e.y -= self.origin[1]
        self.mouse_moved(e)

        return False

    def mouse_event(self, widget, e):
        e.x -= self.origin[0]
        e.y -= self.origin[1]
        self.mouse_pressed(e)

        return False

    def mouse_release_event(self, widget, e):
        e.x -= self.origin[0]
        e.y -= self.origin[1]
        self.mouse_released(e)

        return False

    def key_release_event(self, widget, e):
        self.key_hit(e)

        return False

    def size_allocate_event(self, widget, allocation):
        alloc_size = ((allocation.x, allocation.x + allocation.width),
                      (allocation.y, allocation.y+allocation.height))
        self.resize_canvas(alloc_size)

        return False

    def relax_interaction_timeout(self):
        if self.interactive_level > 0:
            self.interactive_level -= 1
            self.canvas.queue_draw()
        return True
                        

class Toggle:
    def __init__(self):
        self.bounds = (0,0,0,0)
        self.text = ""
        self.on = False
    def draw(self, cr):
        cr.set_source_rgba(0,0,0,1)
        cr.rectangle(bounds.left(), bounds.top(),
                     bounds.width(), bounds.height())
        if(on):
            cr.fill()
            cr.set_source_rgba(1,1,1,1)
        else:
            cr.stroke()
        draw_text(cr, bounds.corner(0) + (5,2), text)
        
    def toggle(self):
        self.on = not self.on
    def set(self, state):
        self.on = state
    def handle_click(self, e):
        if bounds.contains((e.x, e,y)) and e.button == 1:
            toggle()

def toggle_events(toggles, e):
    for t in toggles:
        t.handle_click(e)

def draw_toggles(cr, toggles):
    for t in toggles:
        t.draw(cr)


current_toys = []

def draw_text(cr, loc, txt, bottom = False, font="Sans 12"):
    import pango
    layout = pangocairo.CairoContext.create_layout (cr)
    layout.set_font_description (pango.FontDescription(font))
    
    layout.set_text(txt)
    bounds = layout.get_pixel_extents()
    cr.move_to(loc[0], loc[1])
    if bottom:
        if bottom == "topright":
            cr.move_to(loc[0] - bounds[1][2],
                       loc[1])
        elif bottom == "bottomright":
            cr.move_to(loc[0] - bounds[1][2],
                       loc[1] - bounds[1][3])
        elif bottom == "topleft":
            cr.move_to(loc[0],
                       loc[1])
        else:
            cr.move_to(loc[0], loc[1] - bounds[1][3])
    cr.show_layout(layout)
    return bounds

# Framework Accessors

# Gui Event Callbacks

def make_about(evt):
    about_window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    about_window.set_title("About")
    about_window.set_resizable(True)
    
    about_text = gtk.TextView()
    buf = about_text.get_buffer()
    buf.set_text("Toy lib2geom application", -1)
    about_window.add(about_text)

    about_window.show_all()

def save_cairo(evt):
    d = gtk.FileChooserDialog("Save file as svg, png or pdf",
                              None,
                              gtk.FILE_CHOOSER_ACTION_SAVE,
                              (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                               gtk.STOCK_SAVE, gtk.RESPONSE_ACCEPT))

    if(d.run() == gtk.RESPONSE_ACCEPT):
        filename = d.get_filename()
        cr_s = None
        left, top, width, height = current_toys[0].get_save_size()
        
        if filename[-4:] == ".pdf":
            cr_s = cairo.PDFSurface(filename, width, height)
        elif filename[-4:] == ".png":
            cr_s = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height )
        else:
            cr_s = cairo.SVGSurface(filename, width, height)
        cr = cairo.Context(cr_s)
        cr = pangocairo.CairoContext(cairo.Context(cr_s))
        cr.translate(-left, -top)
        current_toys[0].draw(cr, (width, height), True)
            
        cr.show_page()
        del cr
        del cr_s
    d.destroy()
        
resized = False


def FileMenuAction():
    pass

ui = """
<ui>
  <menubar>
    <menu name="FileMenu" action="FileMenuAction">
      <menuitem name="Save" action="save_cairo" />
      <menuitem name="Quit" action="gtk.main_quit" />
      <placeholder name="FileMenuAdditions" />
    </menu>
    <menu name="HelpMenu" action="HelpMenuAction">
      <menuitem name="About" action="make_about" />
    </menu>
  </menubar>
</ui>
"""

def init(argv, t, width, height):
    global current_toys
    current_toys.append(t)

    t.first_time(argv)
    
    t.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    t.window.set_title("title")

# Creates the menu from the menu data above
    uim = gtk.UIManager()
    ag = gtk.ActionGroup('menu_actiongroup')
    ag.add_actions([('FileMenuAction', None, 'File', None, None, None),
                    ('save_cairo', None, 'Save screenshot', None, None, save_cairo),
                    ('gtk.main_quit', None, 'Quit', None, None, gtk.main_quit),
                    ('HelpMenuAction', None, 'Help', None, None, None),
                    ('make_about', None, 'About', None, None, make_about)
                   ])
    uim.insert_action_group(ag, 0)
    uim.add_ui_from_string(ui)
    menu = uim.get_widget("/ui/menubar")
    # Creates the menu from the menu data above
    #uim = gtk.UIManager()
    #uim.add_ui_from_string(ui)
    #menu = uim.get_widget("ui/menubar")

    t.window.connect("delete_event", t.delete_event)

    t.canvas = gtk.DrawingArea()

    t.canvas.add_events((gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.KEY_PRESS_MASK | gtk.gdk.POINTER_MOTION_MASK | gtk.gdk.SCROLL_MASK))

    t.canvas.connect("expose_event", t.expose_event)
    t.canvas.connect("button_press_event", t.mouse_event)
    t.canvas.connect("button_release_event", t.mouse_release_event)
    t.canvas.connect("motion_notify_event", t.mouse_motion_event)
    t.canvas.connect("key_press_event", t.key_release_event)
    t.canvas.connect("size-allocate", t.size_allocate_event)
    t.canvas.connect("scroll_event", t.scroll_event)
    
    t.vbox = gtk.VBox(False, 0)
    t.window.add(t.vbox)

    t.vbox.pack_start (menu, False, False, 0)

    pain = gtk.VPaned()
    t.vbox.pack_start(pain, True, True, 0)
    pain.add1(t.canvas)

    t.canvas.set_size_request(width, height)
    t.window.show_all()

    # Make sure the canvas can receive key press events.
    t.canvas.set_flags(gtk.CAN_FOCUS)
    t.canvas.grab_focus()
    assert(t.canvas.is_focus())

    t.gtk_ready()
    gobject.timeout_add(1000, t.relax_interaction_timeout)
    
    gtk.main()

def get_vbox():
    return current_toys[0].vbox



#!/usr/bin/python

import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "..", "py2geom"))

from  py2geom import *
import py2geom
import numpy
import random
from py2geom_glue import *

def poly_to_sbasis(p):
    sb = SBasis()
    s = numpy.poly1d([-1, 1, 0])
    while True:
        q,r = p / s
        x = Linear(r[0],r[1]+r[0])
        sb.append(x)
        p = q
        if len(list(p)) <= 1 and p[0] == 0:
            return sb

def sbasis_to_poly(sb):
    p = numpy.poly1d([0])
    s = numpy.poly1d([-1, 1, 0])
    sp = numpy.poly1d([1])
    for sbt in sb:
        p += sp*(sbt[0]*(numpy.poly1d([-1,1])) + sbt[1]*(numpy.poly1d([1,0])))
        sp *= s
    return p

random.seed(1)
trial = numpy.poly1d([random.randrange(0,10) for x in range(6)])

sb = poly_to_sbasis(trial)

pwsb = PiecewiseSBasis()
pwsb.push_seg(sb)
pwsb.push_cut(0)
pwsb.push_cut(1)
print pwsb.size()
print "invariants:", pwsb.invariants()
print pwsb(0)

def l2s(l):
    sb = py2geom.SBasis()
    sb.append(l)
    return sb

X = l2s(py2geom.Linear(0, 1))
OmX = l2s(py2geom.Linear(1, 0))
def bezier_to_sbasis(handles, order):
    print "b2s:", handles, order
    if(order == 0):
        return l2s(py2geom.Linear(handles[0]))
    elif(order == 1):
        return l2s(py2geom.Linear(handles[0], handles[1]))
    else:
        return (py2geom.multiply(OmX, bezier_to_sbasis(handles[:-1], order-1)) +
                py2geom.multiply(X, bezier_to_sbasis(handles[1:], order-1)))


for bz in [[0,1,0], [0,1,2,3]]:
    sb = bezier_to_sbasis(bz, len(bz)-1)
    print bz
    print sb
    print sbasis_to_bezier(sb,0)



#Axis specifiers for Dim2
X = 0
Y = 1

from _cy_primitives import cy_Angle as Angle
from _cy_primitives import cy_AngleInterval as AngleInterval
from _cy_primitives import cy_Point as Point
from _cy_primitives import cy_Line as Line
from _cy_primitives import cy_Ray as Ray
from _cy_primitives import cy_IntPoint as IntPoint


from _cy_rectangle import cy_Interval as Interval
from _cy_rectangle import cy_OptInterval as OptInterval
from _cy_rectangle import cy_IntInterval as IntInterval
from _cy_rectangle import cy_OptIntInterval as OptIntInterval

from _cy_rectangle import cy_GenericInterval as GenericInterval
from _cy_rectangle import cy_GenericOptInterval as GenericOptInterval

from _cy_rectangle import cy_GenericRect as GenericRect

from _cy_rectangle import cy_Rect as Rect
from _cy_rectangle import cy_OptRect as OptRect
from _cy_rectangle import cy_IntRect as IntRect
from _cy_rectangle import cy_OptIntRect as OptIntRect


from _cy_affine import cy_Affine as Affine
from _cy_affine import cy_Translate as Translate
from _cy_affine import cy_Rotate as Rotate
from _cy_affine import cy_VShear as VShear
from _cy_affine import cy_HShear as HShear
from _cy_affine import cy_Scale as Scale
from _cy_affine import cy_Zoom as Zoom

from _cy_affine import cy_Eigen as Eigen

from _cy_curves import cy_Curve as Curve

from _cy_curves import cy_Linear as Linear
from _cy_curves import cy_SBasis as SBasis
from _cy_curves import cy_SBasisCurve as SBasisCurve

from _cy_curves import cy_Bezier as Bezier
from _cy_curves import cy_BezierCurve as BezierCurve
from _cy_curves import cy_LineSegment as LineSegment
from _cy_curves import cy_QuadraticBezier as QuadraticBezier
from _cy_curves import cy_CubicBezier as CubicBezier

from _cy_curves import cy_HLineSegment as HLineSegment
from _cy_curves import cy_VLineSegment as VLineSegment

from _cy_curves import cy_EllipticalArc as EllipticalArc
#Wrap this? It doesn't fit into python's dynamic nature and 
#BezierCurve covers most of it's functionality 
#Maybe implement constructors for BezierCurve similar to those
#seen in BezierCurveN
#TODO
#from _cy_curves import cy_BezierCurveN as BezierCurveN

from _cy_curves import cy_lerp as lerp
from _cy_curves import cy_reverse as reverse
#~ from _cy_curves import cy_level_sets as level_sets


from _cy_path import cy_Path as Path


from _cy_conicsection import cy_Circle as Circle
from _cy_conicsection import cy_Ellipse as Ellipse



from _common_decl cimport *
from cpython.ref cimport PyObject

from _cy_primitives cimport Point, cy_Point, wrap_Point
from _cy_primitives cimport IntPoint, cy_IntPoint, wrap_IntPoint

cdef extern from "2geom/affine.h" namespace "Geom":
    cdef cppclass Affine:
        pass

cdef extern from "2geom/cython-bindings/wrapped-pyobject.h" namespace "Geom":

    cdef cppclass WrappedPyObject:
        WrappedPyObject()
        WrappedPyObject(object)
        object getObj()

    cdef cppclass PyPoint:
        PyPoint()
        PyPoint(WrappedPyObject, WrappedPyObject)
        WrappedPyObject operator[](int)
        

cdef extern from "2geom/generic-interval.h" namespace "Geom":

    cdef cppclass GenericInterval[C]:

        GenericInterval()
        GenericInterval(C)
        GenericInterval(C, C)

        C min()
        C max()
        C extent()
        C middle()
    
        bint isSingular()
        bint contains(C)
        bint contains(GenericInterval[C] &) 
        bint intersects(GenericInterval[C] &)

        void setMin(C)
        void setMax(C)
        void expandTo(C)
        void expandBy(C)
        void unionWith(GenericInterval[C] &)
        
        GenericInterval[C] &operator+(C)
        GenericInterval[C] &operator-(C)
        GenericInterval[C] &operator-()
        GenericInterval[C] &operator+(GenericInterval[C] &)
        GenericInterval[C] &operator-(GenericInterval[C] &)
        GenericInterval[C] &operator|(GenericInterval[C] &)
        bint operator==(GenericInterval[C] &)
        bint operator!=(GenericInterval[C] &)


    cdef cppclass GenericOptInterval[C]:
    
        GenericOptInterval()
        GenericOptInterval(C)
        GenericOptInterval(C, C)

        GenericInterval get() 

        C min()
        C max()
        C extent()
        C middle()

        bint isEmpty()        
        bint isSingular()
        bint contains(C)
        bint contains(GenericOptInterval[C] &) 
        bint intersects(GenericOptInterval[C] &)
        
        void setMin(C)
        void setMax(C)
        void expandTo(C)
        void expandBy(C)
        void unionWith(GenericOptInterval[C] &)
        void intersectWith(GenericOptInterval &)


        GenericOptInterval[C] &operator+(C)
        GenericOptInterval[C] &operator-(C)
        GenericOptInterval[C] &operator-()
        GenericOptInterval[C] &operator+(GenericOptInterval[C] &)
        GenericOptInterval[C] &operator-(GenericOptInterval[C] &)
        GenericOptInterval[C] &operator|(GenericOptInterval[C] &)
        GenericOptInterval[C] &operator&(GenericOptInterval[C] &)
        bint operator==(GenericOptInterval[C] &)
        bint operator!=(GenericOptInterval[C] &)


cdef extern from "2geom/int-interval.h" namespace "Geom":
    ctypedef GenericInterval[IntCoord] IntInterval
    ctypedef GenericOptInterval[IntCoord] OptIntInterval


#redeclaring inherited methods, other option is to cast all
#pointers from Interval to GenericInterval[Coord] in extension class
cdef extern from "2geom/interval.h" namespace "Geom":
    
    cdef cppclass Interval:
    
        Interval()
        Interval(Coord)
        Interval(Coord, Coord)
        
        Coord min()
        Coord max()
        Coord extent()
        Coord middle()
    
        bint isSingular()
        bint contains(Coord)
        bint contains(Interval &) 
        bint intersects(Interval &)
    
        void setMin(Coord)
        void setMax(Coord)
        void expandTo(Coord)
        void expandBy(Coord)
        void unionWith(Interval &)

        bint isFinite()
        bint interiorContains(Coord)
        bint interiorContains(Interval &)
        bint interiorIntersects(Interval &)
        
        Interval & operator*(Coord)
        Interval & operator/(Coord)
        Interval & operator*(Interval &)
        bint operator==(IntInterval &)
        bint operator==(Interval &)
        bint operator!=(Interval &)
        bint operator!=(IntInterval &)                
        Interval &operator+(Coord)
        Interval &operator-(Coord)
        Interval &operator-()
        Interval &operator+(Interval &)
        Interval &operator-(Interval &)
        Interval &operator|(Interval &)
        
        IntInterval roundOutwards()
        OptIntInterval roundInwards()

    
    cdef cppclass OptInterval:

        OptInterval(OptInterval &)
        OptInterval()
        OptInterval(Interval &)
        OptInterval(Coord)
        OptInterval(Coord, Coord)
        OptInterval(GenericOptInterval[double] &)
        OptInterval(IntInterval &)
        OptInterval(OptIntInterval &)
    
        Interval get()
        bint isEmpty()
        void unionWith(OptInterval &)
        void intersectWith(OptInterval &)
        
        OptInterval &operator|(OptInterval &)
        OptInterval &operator&(OptInterval &)

cdef class cy_Interval:
    cdef Interval* thisptr

cdef cy_Interval wrap_Interval(Interval p)

cdef class cy_OptInterval:
    cdef OptInterval* thisptr

cdef cy_OptInterval wrap_OptInterval(OptInterval p)


cdef extern from "2geom/generic-rect.h" namespace "Geom":
    cdef cppclass GenericRect[C]:
        GenericRect()
        GenericRect(GenericInterval[C] &, GenericInterval[C] &)
        GenericRect(Point &, Point &)
        GenericRect(PyPoint &, PyPoint &)
        GenericRect(IntPoint &, IntPoint &)
        GenericRect(C, C, C, C)

        GenericInterval[C] & operator[](Dim2)

        PyPoint min()
        PyPoint max()
        PyPoint corner(unsigned int)
        
        IntPoint i_min "min" ()
        IntPoint i_max "max" ()
        IntPoint i_corner "corner" (unsigned int)
        
        C top()
        C bottom()
        C left()
        C right()
        C width()
        C height()
        Coord aspectRatio()
        
        PyPoint dimensions()
        PyPoint midpoint()

        IntPoint i_dimensions "dimensions" ()
        IntPoint i_midpoint "midpoint" ()
        
        C area()
        bint hasZeroArea()
        C maxExtent()
        C minExtent()
        bint intersects(GenericRect[C] &)
        bint contains(GenericRect[C] &)

        bint contains(PyPoint &)
        bint contains(IntPoint &)

        void setLeft(C)
        void setRight(C)
        void setTop(C)
        void setBottom(C)

        void setMin(PyPoint &)
        void setMax(PyPoint &)
        void expandTo(PyPoint &)
        void setMin(IntPoint &)
        void setMax(IntPoint &)
        void expandTo(IntPoint &)
        
        void unionWith(GenericRect[C] &)
        void expandBy(C)
        void expandBy(C, C)
        
        void expandBy(PyPoint &)
        void expandBy(IntPoint &)
        
        GenericRect[C] & operator+(PyPoint &)
        GenericRect[C] & operator+(IntPoint &)
                
        GenericRect[C] & operator-(PyPoint &)
        GenericRect[C] & operator-(IntPoint &)

        GenericRect[C] & operator|(GenericRect[C] &)
        bint operator==(GenericRect[C] &)
        bint operator!=(GenericRect[C] &)

    
    cdef cppclass GenericOptRect[C]:
        GenericOptRect()
        GenericOptRect(GenericRect[C] &)
        GenericOptRect(C, C, C, C)
        GenericOptRect(Point &, Point &)
        GenericOptRect(GenericOptInterval[C] &, GenericOptInterval[C] &)
        
        GenericRect[C] get()
        
        bint isEmpty()
        bint intersects(GenericRect[C] &)
        bint contains(GenericRect[C] &)
        bint intersects(GenericOptRect[C] &)
        bint contains(GenericOptRect[C] &)
        
        bint contains(Point &)
        bint contains(IntPoint &)
        
        void unionWith(GenericRect[C] &)
        void unionWith(GenericOptRect[C] &)
        void intersectWith(GenericRect[C] &)
        void intersectWith(GenericOptRect[C] &)
        
        void expandTo(Point &)
        
        GenericOptRect[C] &operator|(GenericOptRect[C] &)
        GenericOptRect[C] &operator&(GenericRect[C] &)
        GenericOptRect[C] &operator&(GenericOptRect[C] &)
        
        bint operator==(GenericOptRect[C] &)
        bint operator==(GenericRect[C] &)

        bint operator!=(GenericOptRect[C] &)
        bint operator!=(GenericRect[C] &)

cdef extern from "2geom/generic-rect.h" namespace "Geom::GenericRect<Geom::WrappedPyObject>":
    GenericRect[WrappedPyObject] from_xywh(WrappedPyObject, WrappedPyObject, WrappedPyObject, WrappedPyObject)
    GenericRect[WrappedPyObject] from_xywh(PyPoint &, PyPoint &)

cdef extern from "2geom/int-rect.h" namespace "Geom":
    ctypedef GenericRect[IntCoord] IntRect

cdef extern from "2geom/rect.h" namespace "Geom":
    cdef cppclass Rect:
        Rect()
        Rect(Interval &, Interval &)
        Rect(Point &, Point &)
        Rect(Coord, Coord, Coord, Coord)

        Interval & operator[](Dim2)
        Point min()
        Point max()
        Point corner(unsigned int)
        Coord top()
        Coord bottom()
        Coord left()
        Coord right()
        Coord width()
        Coord height()
        Coord aspectRatio()
        Point dimensions()
        Point midpoint()
        Coord area()
        bint hasZeroArea()
        Coord maxExtent()
        Coord minExtent()

        bint intersects(Rect &)
        bint contains(Rect &)
        bint contains(Point &)

        void setLeft(Coord)
        void setRight(Coord)
        void setTop(Coord)
        void setBottom(Coord)
        void setMin(Point &)
        void setMax(Point &)
        void expandTo(Point &)
        void unionWith(Rect &)
        void expandBy(Coord)
        void expandBy(Coord, Coord)
        void expandBy(Point &)

        Rect & operator+(Point &)
        Rect & operator-(Point &)
        Rect & operator|(Rect &)
        bint operator==(Rect &)
        bint operator!=(Rect &)   
        bint operator==(IntRect &)
        bint operator!=(IntRect &)
        
        bint hasZeroArea(Coord)
        bint interiorIntersects(Rect &)
        bint interiorContains(Point &)
        bint interiorContains(Rect &)
        bint interiorContains(OptRect &)

        IntRect roundOutwards()
        OptIntRect roundInwards()
        Rect &operator*(Affine &)

    Coord distanceSq(Point &, Rect &)
    Coord distance(Point &, Rect &)


    cdef cppclass OptRect:
        OptRect()
        OptRect(Rect &)
        OptRect(Coord, Coord, Coord, Coord)
        OptRect(Point &, Point &)
        OptRect(OptInterval &, OptInterval &)
        
        Rect get()
        
        bint isEmpty()
        bint intersects(Rect &)
        bint contains(Rect &)
        bint intersects(OptRect &)
        bint contains(OptRect &)
        bint contains(Point &)
        
        void unionWith(Rect &)
        void unionWith(OptRect &)
        void intersectWith(Rect &)
        void intersectWith(OptRect &)
        void expandTo(Point &)
        
        OptRect &operator|(OptRect &)
        OptRect &operator&(Rect &)
        OptRect &operator&(OptRect &)
        
        bint operator==(OptRect &)
        bint operator==(Rect &)

        bint operator!=(OptRect &)
        bint operator!=(Rect &)
                
cdef extern from "2geom/rect.h" namespace "Geom::Rect":
    Rect from_xywh(Coord, Coord, Coord, Coord)
    Rect from_xywh(Point &, Point &)
    Rect infinite()

cdef class cy_Rect:
    cdef Rect* thisptr
    
cdef cy_Rect wrap_Rect(Rect p)

cdef class cy_OptRect:
    cdef OptRect* thisptr
    
cdef cy_OptRect wrap_OptRect(OptRect p)


cdef extern from "2geom/int-rect.h" namespace "Geom":
    #redeclaring because cython complains about ambiguous overloading otherwise
    cdef cppclass OptIntRect:
        OptIntRect()
        OptIntRect(IntRect &)
        OptIntRect(IntCoord, IntCoord, IntCoord, IntCoord)
        OptIntRect(IntPoint &, IntPoint &)
        OptIntRect(OptIntInterval &, OptIntInterval &)
        
        IntRect get()
        
        bint isEmpty()
        bint intersects(IntRect &)
        bint contains(IntRect &)
        bint intersects(OptIntRect &)
        bint contains(OptIntRect &)
        
        bint contains(Point &)
        bint contains(IntPoint &)
        
        void unionWith(IntRect &)
        void unionWith(OptIntRect &)
        void intersectWith(IntRect &)
        void intersectWith(OptIntRect &)
        
        void expandTo(IntPoint &)
        
        OptIntRect &operator|(OptIntRect &)
        OptIntRect &operator&(IntRect &)
        OptIntRect &operator&(OptIntRect &)
        
        bint operator==(OptIntRect &)
        bint operator==(IntRect &)

        bint operator!=(OptIntRect &)
        bint operator!=(IntRect &)

